require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'cht-lib-fabric/error_codes'
require 'cht-lib-fabric/crns'

module RoleManagement
  module Actions
    module AuthorizationHelper
      # This method is used to enforce authorization in an action. It must be mised into
      # a class where `context` is available as an local method and that context has
      # an `_request` attribute with the http request being executed. This is
      # required in order to pass the Authorization header on to the authorization
      # service
      #
      # This module must also be mixed into an action that imports the `authorization`
      # client dependency
      #
      # @raise [Platform::GraphQL::Errors::ExecutionError] if the user is not authorized.
      # @return [true] if the user is authorized
      def authorize!(subject:, actions:, require_all: false, context: {})
        return true if service_account_caller?(context)
        enforcement = if require_all
                        Platform::Clients::Authorization::ENFORCEMENT_ALL
                      else
                        Platform::Clients::Authorization::ENFORCEMENT_ANY
                      end

        authorized = authorization.authorized(
          subject: "crn::#{subject}/*",
          actions: Array(actions),
          enforcement: enforcement,
          context: {request: context[:request]}
        )

        raise_error(::Platform::ErrorCodes::NOT_AUTHORIZED, "Not Authorized") unless authorized
        true
      end

      def service_account_caller?(context)
        return @service_account_caller if @service_account_caller

        # if we have been given a valid token (validated by webserver)
        # with from a service account and we can skip authorization
        # and just execute the action
        the_caller = context[:request].jwt_claims[:sub]
        logger.debug "checking for service account for sub #{the_caller}"
        @service_account_caller = begin
                                    subject = subject
                                    ::Platform::CRNs::ServiceAccount.from_s(the_caller)
                                    true
                                  rescue ::Platform::InvalidCRNError
                                    false
                                  end
        logger.debug "is service account?: #{@service_account_caller}"
        @service_account_caller
      end
    end
  end
end