# frozen_string_literal: true

require 'import'
require 'graphql'
require 'cht-lib-fabric/crn'
require 'cht-lib-fabric/tasks/actionable'
require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'role_management/utils/event_publish'

require 'role_management/actions'

module RoleManagement
  module Actions
    class CreateRoleDocument
      include ::Platform::Helpers::GraphQL::RaiseError
      include ::RoleManagement::Actions::AuthorizationHelper
      include Import[
        "logger",
        "role_management.stores.role_documents_directory",
        "role_management.messages.role_document",
        "role_management.utils.event_publish",
        "alert",
        "clients.authorization",
        "role_management.helpers.flow_down_helper"
      ]

      def call (name, description, restricted, permissions, metadata, context)
        authorize!(subject: "role_document", actions: "create", context: context)

        name = name

        if name.to_s.strip.length == 0
          raise_error ::RoleManagement::ErrorCodes::MISSING_ROLE_DOCUMENT_NAME,
                      "Create Role Document: Missing role document name"
        end

        description = description
        permissions = permissions
        restricted = restricted

        customer_id = context[:request].current_customer
        customer_crn = ::Platform::CRNs::Tenant.new(customer: customer_id)
                                               .to_s

        user_id = context[:request].current_user

        user_crn = case user_id
                   # Act as a "nobody" user when invoked by a service account
                   when nil then ::Platform::CRNs::User.new(customer: 0, user: 0).to_s
                   else ::Platform::CRNs::User.new(customer: customer_id,
                                                   user: user_id).to_s
                   end

        organization_id = context[:request].current_organization

        role_document = role_documents_directory.create_custom_role_document(
          name: name,
          description: description,
          restricted: restricted,
          customer_crn: customer_crn,
          user_crn: user_crn,
          permissions: permissions,
          organization: organization_id
        )

        event_publish.publish(role_document, context, ::RoleManagement::Messages::RoleDocument::CREATE)

        role_doc_crn = "crn:#{customer_id}:role-document/#{role_document.id}"
        logger.info "Creating role document #{role_doc_crn}"

        # Create shared role document ids
        fd_resp = flow_down_helper.enable_flow_down(id: role_document.id, context: context, metadata: metadata) unless metadata.nil?

        [role_document, flow_down_helper.to_metadata(fd_response: fd_resp)]
      rescue ::RoleManagement::Exceptions::InvalidPermission => ex
        logger.error "Create Role Document: #{ex.class} #{ex.message}"
        raise_error ::RoleManagement::ErrorCodes::INVALID_PERMISSION,
                    "Create Role Document: invalid permission sent"
      end
    end
  end
end
