# frozen_string_literal: true

require 'import'
require 'graphql'
require 'cht-lib-fabric/crn'
require 'cht-lib-fabric/tasks/actionable'
require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'role_management/errors'
require 'role_management/actions'
require 'role_management/utils/event_publish'

module RoleManagement
  module Actions
    class DeleteRoleDocument
      include ::Platform::Helpers::GraphQL::RaiseError
      include ::RoleManagement::Actions::AuthorizationHelper
      include Import[
        "logger",
        "role_management.stores.role_documents_directory",
        "role_management.messages.role_document",
        "role_management.utils.event_publish",
        "alert",
        "clients.authorization",
        "role_management.helpers.flow_down_helper"
      ]

      def call (id, context)
        authorize!(subject: "role_document", actions: "delete", context:context)

        role_doc_id = ::Platform::CRN.from_s(id)
                                     .details
                                     .to_i

        customer_id = context[:request]&.current_customer
        customer_crn = ::Platform::CRNs::Tenant.new(customer: customer_id)
                                               .to_s
        role_doc_crn = "crn:#{customer_id}:role-document/#{role_doc_id}"

        logger.info "Deleting role document #{role_doc_crn}"

        user_id = context[:request].current_user
        user_crn = ::Platform::CRNs::User.new(customer: customer_id,
                                             user: user_id).to_s

        organization_id = context[:request].current_organization

        # Fetch this so we can publish an event for it, the destroy only
        # returns an ID atm, we really just need the ID for the message, but
        # can't create a roledocument struct w/o alot of the fields
        saved_role_doc = ::RoleManagement::Actions::Node.new.call(::Platform::CRN.from_s(id), context)

        role_document_ar = role_documents_directory.destroy_role_document(role_doc_id,
                                                       customer_crn,
                                                       organization_id)


        event_publish.publish(saved_role_doc, context, ::RoleManagement::Messages::RoleDocument::DELETE)

        # Update shared role document ids
        if flow_down_helper.is_flow_down_item?(crn: ::Platform::CRN.from_s(role_doc_crn), context: { request: context[:request] })
          logger.info "Flow down found for #{role_doc_crn} while deleting role document"
          flow_down_helper.delete_flow_down(id: id, context: context)
        else
          logger.info "No flow down found for #{role_doc_crn} while deleting role document"
        end

        role_doc_crn
      rescue ::ActiveRecord::RecordNotFound
        raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                    "DeleteRoleDocument: role document not found"
      end

    end
  end
end
