# frozen_string_literal: true

require 'import'
require 'graphql'

require 'cht-lib-fabric/crn'
require 'cht-lib-fabric/crns/tenant'
require 'cht-lib-fabric/tasks/actionable'
require 'cht-lib-fabric/helpers/graphql/raise_error'

require 'role_management/errors'
require 'role_management/models/crns'
require 'role_management/actions'

require 'role_management/graphql/include_fields'

module RoleManagement
  module Actions
    class Node
      include ::Platform::Helpers::GraphQL::RaiseError
      include ::RoleManagement::Actions::AuthorizationHelper
      include ::RoleManagement::GraphQL::IncludeFields
      include Import['logger',
                     'role_management.stores.role_documents_directory',
                     'clients.authorization',
                     'role_management.helpers.flow_down_helper',
                     'statsd']

      def call(crn, context)
        if crn.service == 'role-document'
          authorize!(subject: "role_document", actions: "read", context: context)
          fetch_role_document(crn, context)
        elsif crn.service == 'role-mgmt' && /^ugrb/ =~ crn.details
          authorize!(subject: "role_binding", actions: "read", context:context)
          fetch_user_group_role_binding(crn, context)
        else
          raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
          "Node: role document record not found"
        end
      end

      def fetch_role_document(crn, context)
        logger.warn "FETCH ROLE DOCUMENT #{crn} #{context}"
        role_document_id = crn.details.to_i
        crn_customer_id = crn.customer
        customer_id = context[:request].current_customer
        customer_crn = ::Platform::CRNs::Tenant.new(customer: customer_id).to_s
        organization_id = context[:request].current_organization

        include_permissions = include_permissions?(context)
        include_permission_labels = include_permission_labels?(context)

        logger.warn "Checking if role document query includes permissions labels [#{include_permissions}] - [#{include_permission_labels}]"

        role_document = if crn_customer_id > 0
          # Check to see if role doc is present in current org first
           # to avoid making a multiple FDS calls for every custom role doc node query
          begin
            rd = role_documents_directory.by_id_and_customer_crn(
              role_document_id,
              customer_crn,
              organization_id,
              service_account: service_account_caller?(context),
              include_permissions: include_permissions,
              include_permission_labels: include_permission_labels)
            # Check if it was shared in the current org and mark it
            current_org_items = flow_down_helper.get_flow_down_items_by_org(type: crn.service, context: context)
            if current_org_items.include?(role_document_id)
              # Mark item as shared to avoid DB lookup again just to mark
              rd.metadata.shared.attributes.merge!(enabled: true)
            end
            rd
          rescue ::ActiveRecord::RecordNotFound
            logger.warn "Role Document #{crn} not found in current org, checking if it was shared from ancestor org"
            # Check to see if was shared from an ancestory org
            parent_org_items = flow_down_helper.get_flow_down_items(type: crn.service, context: context)
            if parent_org_items.include?(role_document_id)
              # If it is use fetch as a shared role document from a parent org
              role_documents_directory.by_shared_role_doc_id(customer_crn, [role_document_id], include_permission_labels: include_permission_labels)[0]
            else
              # raise so we act accordingly as it was neither found as a shared or
              # non-shared role document
              raise(::ActiveRecord::RecordNotFound)
            end
          end
        else
          role_documents_directory.by_id_and_system_defined(role_document_id, include_permissions: include_permissions, include_permission_labels: include_permission_labels)
        end
        # Record how often we're using the node query to retrieve a shared role document
        unless role_document.nil?
          statsd.increment("role_management.node.fetch_role_document.shared." + (role_document.metadata.shared.enabled ? 'true' : 'false' ))
        end
        role_document
      rescue ::ActiveRecord::RecordNotFound
        raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                    "Node: role document record not found"
      end

      def fetch_user_group_role_binding(crn, context)
        crn = RoleManagement::Models::CRNs::UserGroupRoleBinding.from_s(crn.to_s)
        role_document_id = crn.details.to_i
        crn_customer_id = crn.customer
        customer_id = context[:request].current_customer
        customer_crn = ::Platform::CRNs::Tenant.new(customer: customer_id)
                                               .to_s
        current_ou_crn = ::Platform::CRNs::Organization.new(
          customer: customer_id,
          details: context[:request].current_organization.to_s
        )

        user_group_role_binding = role_documents_directory.user_group_role_binding(
          crn: crn,
          customer_crn: customer_crn,
          organization_crn: current_ou_crn
        )
        user_group_role_binding
      rescue ::RoleManagement::Exceptions::NotFound
        raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                    "Node: user group role binding record not found"
      end
    end
  end
end
