# frozen_string_literal: true

require 'import'
require 'graphql'
require 'securerandom'

require 'cht-lib-fabric/tasks/actionable'

require 'role_management/models'

module RoleManagement
  module Actions
    class PermissionGroups
      include ::RoleManagement::Models::Privileges
      include Import['logger', 'role_management.stores.role_documents_directory']

      def call
        all_privs = all_privileges_struct
        permission_to_crn = role_documents_directory.permission_tuple_to_crn
        all_permissions = Array.new
        all_child_ids = Set.new
        build_hierarchy(all_privs, permission_to_crn, all_permissions, all_child_ids)
        root_permissions(all_permissions, all_child_ids)
      end

      def build_hierarchy(priv_section_or_tuple, permission_to_crn, perms, all_child_ids)
        if priv_section_or_tuple.respond_to?(:children)
          permissions = []
          permission_groups = []
          child_Ids = []
          priv_section_or_tuple.children.each do |child|
            item = build_hierarchy(child, permission_to_crn, perms, all_child_ids)
            next unless item
            if item.is_a?(::RoleManagement::Models::PermissionGroup)
              perms << item
              child_Ids << item.id
              # to maintain a global list of all child elements
              item.childIds.map do |childId|
                all_child_ids << childId
              end
            else
              permissions << item
            end
          end
          # skip partner only permission groups for now
          return nil if priv_section_or_tuple.partner_only
          return ::RoleManagement::Models::PermissionGroup.new(
              groupName: priv_section_or_tuple.label,
              permissionGroup: permission_groups,
              permissions: permissions,
              id: SecureRandom.hex,
              childIds: child_Ids
          )
        else
          return to_permission(priv_section_or_tuple, permission_to_crn)
        end
      end

      def to_permission(priv_tuple, permission_to_crn)
        id  = permission_to_crn[[priv_tuple.action, priv_tuple.subject]]
        if id
          ::RoleManagement::Models::Permission.new(
              id: permission_to_crn[[priv_tuple.action, priv_tuple.subject]],
              action: priv_tuple.action.to_s,
              subject: priv_tuple.subject.to_s,
              label: priv_tuple.label
          )
        else
          # happens if privilege file is more up-to-date then db permissions records
          # and we can resolve the id of the permission
          logger.warn("ID not found for tuple: #{priv_tuple.action}:#{priv_tuple.subject}")
          nil
        end
      end

      # determines if the permission is root permission or not
      def root_permissions(all_permissions, all_child_ids)
        logger.info("Finding Root Permissions among all the permissions")
        permissions = []
        all_permissions.map do |permission|
          if (all_child_ids.exclude?  permission.id)
            permissions << permission.new(isRoot: true)
          else
            permissions << permission
          end
        end
        permissions
      end

    end
  end
end