# frozen_string_literal: true

require 'import'
require 'graphql'
require 'cht-lib-fabric/crn'
require 'cht-lib-fabric/tasks/actionable'
require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'role_management/utils/event_publish'

require 'role_management/actions'

module RoleManagement
  module Actions
    class UpdateRoleDocument
      include ::Platform::Helpers::GraphQL::RaiseError
      include ::RoleManagement::Actions::AuthorizationHelper
      include Import[
        "logger",
        "role_management.stores.role_documents_directory",
        "role_management.messages.role_document",
        "role_management.utils.event_publish",
        "alert",
        "clients.authorization",
        "role_management.helpers.flow_down_helper"
      ]

      def call(id, name, description, restricted, permissions, metadata, context)
        authorize!(subject: "role_document", actions: "update", context:context)

        role_doc_id = id
        name = name

        if name.to_s.strip.length == 0
          raise_error ::RoleManagement::ErrorCodes::MISSING_ROLE_DOCUMENT_NAME,
                      "Create Role Document: Missing role document name"
        end

        description = description
        permissions = permissions

        role_doc_crn = ::Platform::CRN.from_s(role_doc_id)
        id = role_doc_crn.details.to_i

        customer_id = context[:request].current_customer
        customer_crn = ::Platform::CRNs::Tenant.new(customer: customer_id)
                                               .to_s

        logger.info "Updating role document #{role_doc_crn}"                                     

        organization_id = context[:request].current_organization

        if role_doc_crn.customer != customer_id
          raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                      "Update Role Document: role document record not found"
        end

        user_id = context[:request].current_user
        user_crn = ::Platform::CRNs::User.new(customer: customer_id,
                                              user: user_id)
                                          .to_s

        resp = role_documents_directory.update_custom_role_document(id,
                                                                    name: name,
                                                                    description: description,
                                                                    customer_crn: customer_crn,
                                                                    updated_by_crn: user_crn,
                                                                    permissions: permissions,
                                                                    organization: organization_id)

        event_publish.publish(resp, context, ::RoleManagement::Messages::RoleDocument::UPDATE)
                                                                            
        # Update shared role document ids
        fd_resp = flow_down_helper.update_flow_down(id: resp.id, context: context, metadata:metadata) unless metadata.nil?

        [resp, flow_down_helper.to_metadata(fd_response: fd_resp)]
      rescue ::ActiveRecord::RecordNotFound
        raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                    "Update Role Document: role document record not found"
      rescue ::RoleManagement::Exceptions::InvalidPermission => ex
        logger.error "Update Role Document: #{ex.class} #{ex.message}"
        raise_error ::RoleManagement::ErrorCodes::INVALID_PERMISSION,
                    "Update Role Document: invalid permission sent"
      end

    end
  end
end
