# frozen_string_literal: true

require 'import'
require 'graphql'
require 'cht-lib-fabric/crn'
require 'cht-lib-fabric/tasks/actionable'
require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'role_management/utils/event_publish'

require 'role_management/actions'

module RoleManagement
  module Actions
    class UpdateUserGroupRoleBindings
      include ::Platform::Helpers::GraphQL::RaiseError
      include ::RoleManagement::Actions::AuthorizationHelper
      include Import['logger',
                     'role_management.stores.role_documents_directory',
                     'clients.authorization',
                     'role_management.helpers.user_management_helper',
                     'role_management.helpers.org_management_helper',
                     'role_management.messages.role_document',
                     'role_management.utils.event_publish']

      def call (input, context)
        actions = []
        actions << 'create' unless input[:creates].blank?
        actions << 'update' unless input[:updates].blank?
        actions << 'delete' unless input[:deletes].blank?

        authorize!(subject: "role_binding", actions: actions, require_all: true, context: context)

        resp = {}
        role_documents_directory.transaction do
          resp['deletes'] = delete_user_group_role_bindings(input[:deletes], context)
          resp['creates'] = create_user_group_role_bindings(input[:creates], context)
          resp['updates'] = update_user_group_role_bindings(input[:updates], context)
        end
        resp
      end

      def create_user_group_role_bindings(creates, context)
        return nil if creates.nil?
        logger.debug("creating user group role bindings : #{creates} ")
        customer_crn, user_crn, current_ou_crn = context_info(context)

        creates.map do |ugrb|
          begin
            valid_user_group?(context, ugrb[:user_group][:id], ugrb[:role_document][:id])
            valid_organizations?(context, ugrb[:organizations].map{|x| x[:id]})

            role_document_crn = ::RoleManagement::Models::CRNs::RoleDocument.from_s(ugrb[:role_document][:id])
            role_binding = role_documents_directory.create_user_group_role_binding(
              customer_crn: customer_crn,
              user_crn: user_crn,
              current_ou_crn: current_ou_crn,
              user_group_crn: ugrb[:user_group][:id],
              role_document_crn: role_document_crn,
              organization_crns: ugrb[:organizations].map{|x| x[:id]}
            )
            event_publish.publish(role_binding, context, ::RoleManagement::Messages::RoleBinding::CREATE)
            role_binding
          rescue ::Platform::InvalidCRNError
            raise_error ::RoleManagement::ErrorCodes::INVALID_CRN,
                        "Invalid role document crn #{role_document_crn.to_s}"
          rescue ::RoleManagement::Exceptions::InvalidRoleDocument
            raise_error ::RoleManagement::ErrorCodes::INVALID_ROLE_DOCUMENT,
            "Role document with crn #{role_document_crn.to_s} does not exist"
          rescue ::RoleManagement::Exceptions::BindingAlreadyExists
            raise_error ::RoleManagement::ErrorCodes::BINDING_ALREADY_EXISTS,
            "User group role binding already exists"
          end
        end
      end

      def update_user_group_role_bindings(updates, context)
        return nil if updates.nil?
        logger.debug("updating user group role bindings : #{updates} ")
        customer_crn, user_crn, current_ou_crn = context_info(context)
        updates.map do |ugrb|
          valid_organizations?(context, ugrb[:organization_ids])
          crn = RoleManagement::Models::CRNs::UserGroupRoleBinding.from_s(ugrb[:id])
          role_binding = role_documents_directory.update_user_group_role_binding(
            customer_crn: customer_crn,
            user_crn: user_crn,
            current_ou_crn: current_ou_crn,
            ugrb_crn: crn,
            organization_crns: ugrb[:organization_ids]
          )
          event_publish.publish(role_binding, context, ::RoleManagement::Messages::RoleBinding::UPDATE)
          role_binding
        end
      rescue ::RoleManagement::Exceptions::NotFound
        raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                    "Role document record not found"
      end

      def delete_user_group_role_bindings(deletes, context)
        return nil if deletes.nil?
        logger.debug("Deleting user group role bindings : #{deletes} ")
        customer_crn, user_crn, current_ou_crn = context_info(context)
        deletes.map do |ugrb|
          crn = RoleManagement::Models::CRNs::UserGroupRoleBinding.from_s(ugrb[:id])
          # Save this off temporarily as the delete only returns an ID
          saved_binding = role_documents_directory.user_group_role_binding(crn: crn, 
            customer_crn: customer_crn,
            organization_crn: current_ou_crn)
          role_binding = role_documents_directory.delete_user_group_role_binding(
            customer_crn: customer_crn,
            user_crn: user_crn,
            current_ou_crn: current_ou_crn,
            ugrb_crn: crn,
          )
          event_publish.publish(saved_binding, context, ::RoleManagement::Messages::RoleBinding::DELETE)
          role_binding
        end
      rescue ::RoleManagement::Exceptions::NotFound
        raise_error ::RoleManagement::ErrorCodes::NOT_FOUND,
                    "Role document record not found"
      end

      private

      def context_info(context)
        customer_id = context[:request].current_customer
        customer_crn = ::Platform::CRNs::Tenant.new(customer: customer_id)

        user_id = context[:request].current_user
        user_crn = case user_id
                   # Act as a "nobody" user when invoked by a service account
                   when nil then ::Platform::CRNs::User.new(customer: 0, user: 0)
                   else ::Platform::CRNs::User.new(customer: customer_id, user: user_id)
                   end
        current_ou_crn = ::Platform::CRNs::Organization.new(
          customer: customer_id,
          details: context[:request].current_organization.to_s
        )

        [customer_crn, user_crn, current_ou_crn]
      end

      #validates if organization exists by calling org-management-service
      def valid_organizations?(context, orgs)
        logger.debug("checking organizations validity")
        invalid_orgs = []

        orgs.each do |org|
          organization = org_management_helper.get_organization(context, org.to_s)
          if (organization.nil?)
            invalid_orgs << org
          end
        end

        if(invalid_orgs.count != 0)
          raise_error ::RoleManagement::ErrorCodes::INVALID_ORGANIZATION, "invalid organizations : #{invalid_orgs}"
        end

        true
      end

      #validates if user-group exists by calling user-management-service
      #if the user group contains a SIA account user, it also validates that the
      #user group is being bound to the same role document as in its configuration
      def valid_user_group?(context, user_group_crn, role_document_id)
        logger.debug("checking #{user_group_crn} validity")

        user_group = user_management_helper.get_user_group(context[:request], user_group_crn)

        if (user_group.nil?)
          raise_error ::RoleManagement::ErrorCodes::INVALID_USER_GROUP, "User group #{user_group_crn.to_s} does not exist"
        else
          # Do some additional validations
          user_management_helper.validate_user_group?(context[:request], user_group, role_document_id, user_group_crn)
        end

        true
      end

    end
  end
end
