# frozen_string_literal: true

require 'import'
require 'cht-lib-fabric/tasks/actionable'
require 'role_management/actions'

module RoleManagement
  module Actions
    # Cleanup user group role bindings related to user group that has been deleted.
    #
    # The input is an instance of `Platform::Tasks::Actionable::Context`
    # with an attribute `:crn` which is a `Platform::CRN::UserGroup`
    # object that identifies the user group that has been deleted.
    class UserGroupDeleted
      include ::Platform::Tasks::Actionable
      include Import['logger',
                     'role_management.stores.role_documents_directory',
                     'role_management.utils.event_publish']

      def call
        crn = context.crn
        customer_id = context.customer_id
        organization_id = context.organization_id
        user_id = context.user_id

        # Manufacture a context for passing as we dont have one here
        ctx = get_context_for_message(customer_id, user_id, organization_id)

        logger.info { "Cleaning up user group role bindings for user-group #{crn}" }
        user_group_role_bindings = role_documents_directory.user_group_role_bindings_by_user_group(user_group_crn: crn)

        user_group_role_bindings.each do |ugrb|
          role_documents_directory.delete_user_group_role_binding(
              customer_crn: ugrb.customer_id,
              user_crn: ugrb.created_by_id,
              current_ou_crn: ugrb.organization_id,
              ugrb_crn: RoleManagement::Models::CRNs::UserGroupRoleBinding.from_s(ugrb.id))

          event_publish.publish(ugrb, ctx, ::RoleManagement::Messages::RoleBinding::DELETE)
        end
      end

      #
      # Manufacture a request context, this request comes in
      # through a REST endpoint, the majority of publishers run inside
      # a graphql mutation which have a request context provided
      #
      # @param - customer_id     - The integer value of a customer ID
      # @param - user_id         - The integer value of a user ID
      # @param - organization_id - The integer value of an organization ID
      def get_context_for_message(customer_id, user_id, organization_id)
        req = Platform::Contexts::Request.new
        req.current_customer = customer_id
        req.current_organization = organization_id
        req.current_user = user_id
        { :request => req }
      end

    end
  end
end
