# frozen_string_literal: true

require 'cht-lib-fabric/graphql/federation/types'
require 'cht-lib-fabric/helpers/graphql/raise_error'
require 'cht-lib-fabric/helpers/graphql/transaction_resolver'

module RoleManagement
  module GraphQL
    module Types
      class Query < ::Platform::GraphQL::Types::BaseObject
        include Import['logger']

        description "Query root"

        add_field(::GraphQL::Types::Relay::NodeField)

        field :permission_groups, [PermissionGroup], null: false do
          description "Get all permission groups"
        end

        def permission_groups
          logger.info "Fetching permission groups"
          permission_groups = ::RoleManagement::Actions::PermissionGroups.new.call
          permission_groups
        end

        field :permission_hierarchy, PermissionGroup, null: false do
          description "Get permission hierarchy"
        end

        def permission_hierarchy
          logger.info "Fetching permission hierarchy"
          permission_hierarchy = ::RoleManagement::Actions::PermissionHierarchy.new.call
          permission_hierarchy
        end

        field :role_documents, RoleDocument.connection_type, null: false,  extras: [:lookahead] do
          description "Get all role documents within the current authorization context"
        end

        def role_documents(lookahead:)
          logger.info "Fetching role documents"
          include_permissions = lookahead.selection(:nodes).selects?(:permissions) ||
              lookahead.selection(:edges).selection(:node).selects?(:permissions)
          role_documents = ::RoleManagement::Actions::RoleDocuments.new.call(context, include_permissions: include_permissions)
          role_documents
        end

        field :user_group_role_bindings, UserGroupRoleBinding.connection_type, null: false, extras: [:lookahead] do
          description "Get all user group role bindings within the current authorization context"
        end

        def user_group_role_bindings(lookahead:)
          logger.info "Fetching user group role bindings"
          role_documents = ::RoleManagement::Actions::UserGroupRoleBindings.new.call(context, lookahead: lookahead)
          role_documents
        end

        field :shared_role_document, SharedRoleDocument, null: false do
          argument :id, ID, required: true
          description "Gets all user groups and organizations for a shared role document"
        end

        def shared_role_document(**args)
          logger.info "Fetching shared role document"
          shared_role_document = shared_role.call(self, args, context)
          shared_role_document
        end

        private

        def shared_role
          @shared_role ||= Platform::Helpers::GraphQL::TransactionResolver.new(
              Application, "role_management.transactions.shared_role_document"
          )
        end

      end
    end
  end
end