# frozen_string_literal: true

require 'container'
require 'role_management/actions'
require 'role_management/graph_ql/types'
require 'cht-lib-fabric/helpers/graphql/raise_error'

require 'cht-lib-fabric/graphql/federation/types'
require 'cht-lib-fabric/helpers/graphql/metric_tracing'
require 'cht-lib-fabric/helpers/graphql/id/coerce_input'

module RoleManagement
  module GraphQL
    class Schema < ::GraphQL::Schema
      extend ::Platform::Helpers::GraphQL::RaiseError
      extend ActiveSupport::Inflector
      include ApolloFederation::Schema

      use ::GraphQL::Pagination::Connections
      use ::Platform::Helpers::GraphQL::MetricTracing,
          client: Application[:wavefront]

      # currently setting as 17 as one of the apitest used for testing
      max_depth 17

      query Types::Query
      mutation Types::Mutation

      orphan_types [
                       Types::RoleDocument,
                       Types::UserGroupRoleBinding
                   ]

      def self.resolve_type(_type, obj, _ctx)
        class_basename = demodulize(obj.class.name)
        begin
          Types.const_get(class_basename)
        rescue NameError => ex
          logger.error(ex)
          raise_error(
              ::Platform::ErrorCodes::INTERNAL_ERROR,
              "Type #{class_basename} does not have a corresponding type class"
          )
        end
      end

      def self.id_from_object(object, _type, context)
      end

      def self.logger
        Application["logger"]
      end

      def self.object_from_id(unique_id, context)
        begin
          crn = ::Platform::CRN.from_s(unique_id)
        rescue ::Platform::InvalidCRNError
          raise_error ::RoleManagement::ErrorCodes::INVALID_CRN,
                      "Invalid CRN format with query"
        end

        if crn.customer > 0 && context[:request].current_customer != crn.customer
          raise_error(
              ::Platform::ErrorCodes::NOT_FOUND,
              "Tenant in CRN NOT owned by requested user"
          )
        end

        if !((crn.service == 'role-document') || (crn.service == 'role-mgmt' && /^ugrb/ =~ crn.details))
          raise_error(
              ::Platform::ErrorCodes::INVALID_CRN_SERVICE,
              "#{crn.service} is not a valid service name"
          )
        end

        result = ::RoleManagement::Actions::Node.new.call(crn, context)
        result
      end

    end
  end
end
