# frozen_string_literal: true

require 'import'
require 'cht-lib-fabric/crn'

module RoleManagement
  module Helpers
    class FlowDownHelper
      include Import['logger',
                     'clients.flowdown',
                     'alert']

      ITEM_ALREADY_EXISTS = "EnableFlowDown: flow down item already exists"
      ITEM_DOESNT_EXIST   = "EnableFlowDown: flow down item not found"
      #
      # Returns an array of IDS that represent flowdown
      # items
      #
      def get_flow_down_items(type:, context:)
        logger.debug "FlowDownHelper::get_flow_down_items"
        # Fetch shared role documents
        begin
          ids = flowdown.flow_down_items(type: type, context: context)
          get_flow_down_ids(flow_down_ids: ids)
        rescue => ex
          # Log/Rollbar but don't re-raise, the end result is the same
          # we don't want to fail the entire operation if flow down
          # is unavailable.
          logger.error "Error occuring getting flow down items: #{type} - #{ex.class} - #{ex.message}"
          alert.exception(ex, "FlowDownHelper::get_flow_down_items")
          []
        end
      end

      #
      # Returns an array of IDS that represent flowdown
      # items for the current organization
      #
      def get_flow_down_items_by_org(type:, context:)
        logger.debug "FlowDownHelper::get_flow_down_items_by_org"
        # Fetch shared role documents
        begin
          ids = flowdown.flow_down_items_by_org(type: type, context: context)
          get_flow_down_ids(flow_down_ids: ids)
        rescue => ex
          # Log/Rollbar but don't re-raise, the end result is the same
          # we don't want to fail the entire operation if flow down
          # is unavailable.
          logger.error "Error occuring getting flow down items by organization: #{type} - #{ex.class} - #{ex.message}"
          alert.exception(ex, "FlowDownHelper::get_flow_down_items_by_org")
          []
        end        
      end

      #
      # Enables flow down for an item
      #
      def enable_flow_down(id:, context:, metadata:)
        request = context[:request]

        begin
          if metadata["shared"]["enabled"]
            # Call out to enable flow down
            flowdown.enable_flow_down(id: id, enabled: true, context: { request: request })
          end
        rescue => ex
          # Catch other exceptions here
          begin
            handle_exception(ex: ex)
          rescue => ex
            logger.error "Error occuring enabling flowdown: #{id} - #{ex.class} - #{ex.message}"
            alert.exception(ex, "FlowDownHelper::enable_flow_down")
          end
          reform_response(id: id, metadata: metadata)
        end
      end

      #
      # Disables flow down for an item
      #
      def update_flow_down(id:, context:, metadata:)
        request = context[:request]
        begin
          if metadata["shared"]["enabled"]
            # Call out to enable flow down
            flowdown.enable_flow_down(id: id, enabled: true, context: { request: request })
          else
            # Call out to disable flow down
            flowdown.enable_flow_down(id: id, enabled: false, context: { request: request })
          end
        rescue => ex
          begin
            handle_exception(ex: ex)
          rescue => ex
            # we don't want to fail the entire operation if flow down
            # is unavailable.
            logger.error "Error occuring updating flow down item: #{id} - #{ex.class} - #{ex.message}"
            alert.exception(ex, "FlowDownHelper::update_flow_down")
          end
          reform_response(id: id, metadata: metadata)
        end
      end

      #
      # Deletes flow down for an item
      #
      def delete_flow_down(id:, context:)

        begin
          flowdown.enable_flow_down(id: id, enabled: false, context: { request: context[:request] })
        rescue => ex
          handle_exception(ex: ex)
        end
      end

      def handle_exception(ex:)
        # Flow Down currently enforces you can not enable flow down for
        # an item that is already enabled, and you can't disable for
        # an item that isnt enabled. This presents a challenge for update(s)
        # that can not tell if an item is having sharing enabled for the first time or
        # is an update to the item (in this case createRoleDocument vs updateRoleDocument)
        #
        # This is a HACK in place until we can have RMS call out to the
        # Flow Down Service and check for an item in the scope of the current
        # org.  FDS supports it but by default only checks for an item in the
        # context of ancestor orgs. So we'll NOOP it here with a warning for now.
        if ex.message.include?(ITEM_ALREADY_EXISTS) || ex.message.include?(ITEM_DOESNT_EXIST)
          logger.warn "Item either didnt exist or was already enabled when updating sharing"
        else
          # Still throw other kinds of exceptions
          raise ex
        end
      end

      #
      # Returns a formatted meta data response object
      #
      def to_metadata(fd_response:)
        # {"data":{"items":[{"id":"crn:2017:role-document/4","enabled":true}]},"casted_data":{},"errors":[]}
        enabled = false
        unless fd_response.nil?
          logger.debug "to_metadata - flow down response: #{fd_response.to_h.with_indifferent_access}"
          response = fd_response.to_h.with_indifferent_access
          enabled = response['items'][0]['enabled'] unless response['items'].nil?
        end
        ::RoleManagement::Models::Content::Metadata.new(
          shared: ::RoleManagement::Models::Content::SharedContentMetadata.new(
            enabled: enabled
          )
        )
      end

      #
      # Returns a boolean indicating whether a CRN represnts a flow down item
      #
      def is_flow_down_item?(crn:, context:)
        # Check in the context of items shared with our org and by an org
        items = []
        items_shared_by_org = get_flow_down_items_by_org(type: crn.service, context: context)
        items_shared_with_me = get_flow_down_items(type: crn.service, context: context)
        items = items_shared_by_org + items_shared_with_me
        logger.debug "is_flow_down_item? : items = #{items}"
        return false if items.empty?
        items.include?(crn.details.to_i)
      end

      # This is to reconsitute a responsee until
      # ENG-23216 and ENG-23073 are both in
      def reform_response(id:, metadata:)
        {
          items: [{ id: id, enabled: metadata["shared"]["enabled"] }]
        }
      end

      #
      # Extract the flow down ids
      #
      def get_flow_down_ids(flow_down_ids: [])
        ids = []
        flow_down_ids&.each do |id|
          flow_down_crn = ::Platform::CRN.from_s(id.to_h.with_indifferent_access[:id])
          logger.debug "Flow Down CRN [#{flow_down_crn.to_json}]"
          ids << flow_down_crn.details.to_i
        end

        logger.debug "Flow Down IDs [#{ids}]"
        ids
      end
    end
  end
end
