# frozen_string_literal: true

require 'import'
require 'cht-lib-fabric/helpers/graphql/raise_error'

module RoleManagement
  module Helpers
    class UserManagementHelper
      include ::Platform::Helpers::GraphQL::RaiseError
      include Import['logger',
                     'clients.user_management']

      #
      # Returns all user groups for the given customerId
      #
      def get_user_groups(context_request)
        logger.info "Fetching user groups from user management service"
        customer_id = context_request.current_customer
        user_groups = user_management.get_user_groups(request: context_request)
        if user_groups.nil?
          logger.debug "Unable to fetch user groups for the customer : #{customer_id}"
          nil
        else
          logger.debug "UserGroup fetched for customer : #{customer_id}"
          user_groups
        end
      end

      def get_user_group(context_request, user_group_id)
        logger.info "Fetching user group from user management service"
        customer_id = context_request.current_customer
        user_group = user_management.get_user_group(request: context_request, id: user_group_id)
        if user_group.nil?
          logger.debug "Unable to fetch user group for the customer : #{customer_id}"
          nil
        else
          logger.debug "UserGroup fetched for customer : #{customer_id}"
          user_group
        end
      end

      def get_service_integration_accounts(context_request)
        logger.info "Fetching service_integration accounts from user management service"
        customer_id = context_request.current_customer
        sia_accounts = user_management.get_service_integration_accounts(request: context_request)
        if sia_accounts.nil?
          logger.debug "Unable to fetch service integration for the customer : #{customer_id}"
          nil
        else
          logger.debug "Service Integration Accounts fetched for customer : #{customer_id}"
          sia_accounts
        end
      end

      def lookup_service_integration(id, service_integrations)
        service_integrations.dig(:serviceIntegrationUsers, :edges)&.each do |service_integration|
          val = app_key = service_integration.keys[0]
          app = service_integration[app_key]
          logger.debug "Comparing id #{id} to app id #{app['id']} - #{app['id'].to_s == id.to_s }"
          return app if app['id'].to_s == id.to_s
        end
        logger.error("Unknown service integration id provided #{id}")
        raise ::GraphQL::ExecutionError, "Unknown service integration id provided #{id}"
      end

      # Perform some additional validations if the user group
      # has a service integration account. In the case of any
      # error we'll present the same error as if the group was
      # invalid back to the caller.
      def validate_user_group?(context, user_group, role_document_id, user_group_crn)
        # Should be exactly one SIA account for a group, so we'll use 0
        user = user_group.dig(:node, :users, :edges)&.dig(0)&.dig(:node)
        logger.debug("Found SIA user [#{user.to_json}]")
        if user && user[:serviceAccount]
          # If we determine its a user group with a SIA user in it, then we
          # call out to get service account config
          service_integration_accounts = get_service_integration_accounts(context)
          logger.debug("Configured SIA Accounts [#{service_integration_accounts.to_json}]")
          if service_integration_accounts
            sia_crn = "crn::service-integration/#{user[:oauthAppId]}"
            sia_account_config = lookup_service_integration(sia_crn, service_integration_accounts)
            logger.debug("Comparing [#{sia_account_config[:roleDocumentId].to_s}] to [#{role_document_id.to_s}]")
            if sia_account_config && (sia_account_config[:roleDocumentId].to_s == role_document_id.to_s)
              true
            else
              raise_error ::RoleManagement::ErrorCodes::INVALID_USER_GROUP, "User group #{user_group_crn.to_s} may only be bound to its configured integration account role document" 
            end
          else
            raise_error ::RoleManagement::ErrorCodes::INVALID_USER_GROUP, "User group #{user_group_crn.to_s} could not be matched to its integration account"
          end
        end
        # if no users (or its a regular user) then we just treat it like a regular user group
        true
      end

    end
  end
end