# Representation of the legacy privilege model for seeding join data
# between builtin roles and privileges
module RoleManagement
  module Models
    module Privileges
      REPORT_PRIVILEGES = [:read, :export, :save, :share, :subscribe]
      ASSET_PRIVILEGES = [:read, :export, :save, :subscribe]
      TAGGABLE_ASSET_PRIVILEGES = ASSET_PRIVILEGES + [:tag]
      CRUD_PRIVILEGES = [:create, :read, :update, :delete]
      UDC_PRIVILEGES = [:read_owned, :update_owned, :delete_owned]
      AZURE_PARTNER_CUSTOMER_ACCOUNT_PRIVILEGES = [:create, :export, :read, :save, :subscribe]
      OCI_GENERIC_ASSET_PRIVILEGES = [:read, :save, :subscribe, :tag]
      OCI_REGION_PRIVILEGES = [:export, :read, :save, :subscribe]
      ALICLOUD_GENERIC_ASSET_PRIVILEGES = [:read, :save, :subscribe, :tag, :export, :share]
      DATA_CONNECT_PRIVILEGES = [:scan, :save, :subscribe]
      GCP_GCE_USAGE_REPORT_PRIVILEGES = [:read, :export, :share, :subscribe, :read_owned]
      GCP_BQ_USAGE_REPORT_PRIVILEGES = [:read, :export, :share, :subscribe, :read_owned]
      GCP_BQ_COST_REPORT_PRIVILEGES = [:read, :export, :share, :subscribe, :read_owned]
      K8S_REPORT_PRIVILEGES = [:read, :export, :share, :subscribe, :read_owned]

      PrivilegeSection = Struct.new(:original_label, :hidden, :children, :partner_only) do
        def unique_id
          original_label.underscore.gsub(/\W/, '-')
        end

        def label
          @key ||= "PrivilegeSection.#{unique_id}"
          Sgtn.t(@key, 'cht-svc-role-management')
        end

        def child_privileges
          children.inject([]) do |memo, priv|
            case
            when priv.hidden then memo
            when priv.is_a?(PrivilegeTuple) then memo + [priv]
            when priv.is_a?(PrivilegeSection) then memo + priv.child_privileges
            end
          end
        end
      end

      PrivilegeTuple = Struct.new(:action, :subject, :original_label, :hidden) do
        def initialize(action, subject, original_label = nil, hidden = false)
          super
          ensure_label
        end

        def label
          @key ||= "PrivilegeTuple.#{unique_id}"
          Sgtn.t(@key, 'cht-svc-role-management')
        end

        def ensure_label
          return unless original_label.nil?
          label_action = action.to_s.end_with?('_owned') ? action.to_s.gsub('_owned', '_own') : action
          self.original_label = acronymize("#{label_action.to_s.humanize.titleize} #{subject.to_s.humanize.titleize}")
        end

        def unique_id
          "#{action}_#{subject}"
        end

        def to_sym
          unique_id.to_sym
        end

        private

        def acronymize(string)
          string.gsub(/\b(aws|gcp|rds|ri|ec2|api)\b/i){|m| m.upcase}
        end
      end

      STANDARD_USER_ALLOWLIST_ACTIONS = [
          :read,
          :read_owned,
          :save,
          :share,
          :export,
          :subscribe
      ]

      STANDARD_USER_ALLOWLIST_PRIVILEGES = [
        [:create, :viz_schedule],
        [:create, :viz_preference],
        [:create, :custom_report],
        [:create, :custom_report_subscription],
        [:create, :export_job],
        [:create, :azure_sql_db_rightsizing_reports],
        [:create, :webhook],
        [:update, :webhook],
        [:update, :export_job],
        [:update_owned, :viz_schedule],
        [:update_owned, :viz_preference],
        [:update_owned, :custom_report],
        [:update_owned, :custom_report_subscription],
        [:delete_owned, :viz_schedule],
        [:delete_owned, :viz_preference],
        [:delete_owned, :custom_report],
        [:delete_owned, :custom_report_subscription],
        [:self_edit, :user]
      ]

      STANDARD_USER_DENYLIST_PRIVILEGES = [
        [:read, :organization],
        [:read_owned, :organization],
        [:read, :service_integration],
        [:approve, :service_integration],
        [:revoke, :service_integration]
      ]

      # TMC integration Privileges
      TMC_ALLOWLIST_PRIVILEGES = [
        [:read, :flex_report],
	[:read, :aws_cur_flex_dataset],
	[:read, :azure_ea_flex_dataset],
	[:read, :aws_kubernetes_cost_dataset],
	[:read, :azure_kubernetes_cost_dataset]
      ]

      # each entry is [action, subject]
      def all_privileges
        @all_privileges ||= privilege_nodes(all_privileges_struct)
          .map(&:to_a)
          .select{|v| v.kind_of?(Array)}
          .uniq
          .map{|p| p[0..1]}
      end

      def all_customer_privileges
        @all_customer_privileges ||= privilege_nodes(all_customer_privileges_struct)
          .map(&:to_a)
          .select{|v| v.kind_of?(Array)}
          .uniq
          .map{|p| p[0..1]}
      end

      def builtin_admin_privileges
        all_privileges
      end

      def builtin_power_user_privileges
        denylist = [
          generate_crud_privileges(:organization),
          generate_crud_privileges(:flex_org),
          [
            PrivilegeTuple.new(:self_edit_organization, :user),
            PrivilegeTuple.new(:self_edit_roles, :user),
            PrivilegeTuple.new(:create, :user),
            PrivilegeTuple.new(:update, :user),
            PrivilegeTuple.new(:delete, :user),
            PrivilegeTuple.new(:bulk_invite, :user),
            PrivilegeTuple.new(:create, :role),
            PrivilegeTuple.new(:update, :role),
            PrivilegeTuple.new(:delete, :role),
            PrivilegeTuple.new(:read, :service_integration),
            PrivilegeTuple.new(:approve, :service_integration),
            PrivilegeTuple.new(:revoke, :service_integration),
          ]
        ].flatten(1)
          .uniq
          .map{|p| p.values_at(0..1)}

        all_customer_privileges - denylist
      end

      def builtin_customer_admin_privileges
        all_customer_privileges
      end

      def builtin_standard_user_privileges
        all_customer_privileges.select do |priv|
          action = priv.first
          (STANDARD_USER_ALLOWLIST_ACTIONS.include?(action) || STANDARD_USER_ALLOWLIST_PRIVILEGES.include?(priv)) &&
            !STANDARD_USER_DENYLIST_PRIVILEGES.include?(priv)
        end
      end

      # Aria Hub Integraion Role Document Privileges
      def builtin_tmc_user_privileges
        all_customer_privileges.select do |priv|
          TMC_ALLOWLIST_PRIVILEGES.include?(priv)
        end
      end

      def privilege_object(label, privs, hidden=false, partner_only=false)
        PrivilegeSection.new(label, hidden, privs.uniq, partner_only)
      end

      def core_privilege_objects
        [
          privilege_object('Customer', owned_customer_privileges),
          privilege_object('AWS', aws_privileges),
          privilege_object('Google', gcp_privileges),
          privilege_object('VMWare', vmware_privileges, true),
          privilege_object('Data Center', data_center_privileges),
          privilege_object('Azure', azure_privileges),
          privilege_object('Oracle', oracle_privileges),
          privilege_object('VMware Cloud', vmware_cloud_privileges, hide_feature?('vmc_integration')),
          privilege_object('Multi Cloud', multi_cloud_privileges),
          privilege_object('Alibaba', alicloud_privileges),
          privilege_object('Kubernetes', kubernetes_privileges, hide_feature?('containers_flexorgs')),
          privilege_object('Saved Reports', saved_report_privileges),
          privilege_object('Multidimensional Reports', multidimensional_reports_privileges),
          privilege_object('FlexReports', flex_report_privileges),
          privilege_object('Statements', generate_report_privileges(:statement)),
          privilege_object('Setup', setup_privileges),
          privilege_object('Integrations', integration_privileges),
          privilege_object('Services', services_privileges),
          privilege_object('Containers', container_report_privileges),
          privilege_object('Recommendations', recommendation_privileges),
          privilege_object('Other', other_privileges),
          privilege_object('Partner', partner_privileges, hidden = false, partner_only = true)
        ]
      end

      def all_privileges_struct
        privilege_objects = core_privilege_objects
        privilege_objects << privilege_object('Create Customer', create_customer_privilege, true)
        all_privs = PrivilegeSection.new("All Privileges", false, privilege_objects)
        mark_hidden_privileges(all_privs)
        all_privs
      end

      def all_customer_privileges_struct
        privilege_objects = core_privilege_objects
        privilege_objects << privilege_object('Create Customer', create_customer_privilege, true)
        all_privs = PrivilegeSection.new("All Privileges", false, privilege_objects)
        mark_hidden_privileges(all_privs)
        all_privs
      end

      ##
      # For various reasons there are a number of individual privileges
      # we may want to hide at any given time. This code will go through
      # the hierarchy and mark them as hidden.
      #
      # An example of a hidden privilege is editing your own budget. We
      # don't have the concept of multiple budgets yet, so it makes no
      # sense to own one.
      HIDDEN_PRIVILEGES = [
        [:read_owned, :perspective],
        [:update_owned, :perspective],
        [:delete_owned, :perspective],
        [:read_owned, :budget],
        [:update_owned, :budget],
        [:delete_owned, :budget],
        [:read_owned, :dashboard],
        [:update_owned, :dashboard],
        [:delete_owned, :dashboard],
      ]
      def mark_hidden_privileges(privs)
        visit_privilege_nodes(privs) do |priv|
          priv.hidden = true if HIDDEN_PRIVILEGES.include?([priv.action, priv.subject])
        end
      end

      #############################################
      # Cloud Agnostic Privileges
      #############################################
      def setup_privileges
        [
          privilege_object('Cost Drilldowns', ir_drilldown_cost_privileges),
          privilege_object('Aggregators', aggregator_privileges),
          privilege_object('Organizations', generate_crud_privileges(:organization)),
          privilege_object('FlexOrgs', generate_crud_privileges(:flex_org)),
          privilege_object('Perspectives', generate_udc_privileges(:perspective)),
          privilege_object('Governance', governance_privileges),
          privilege_object('Profile', profile_privileges),
          privilege_object('Roles', generate_crud_privileges(:role)),
          privilege_object('Role Documents',
                           generate_crud_privileges(:role_document),
                           hide_feature?('enterprise_edition')),
          privilege_object('Role Bindings',
                           generate_crud_privileges(:role_binding),
                           hide_feature?('enterprise_edition')),
          privilege_object('Users', users_privileges),
          privilege_object('User Groups', generate_crud_privileges(:user_group),
                           hide_feature?('enterprise_edition')),
          privilege_object('User Group IdP Mappings', generate_crud_privileges(:user_group_idp_mapping),
                           hide_feature?('enterprise_edition')),
          privilege_object('Containers', containers_privileges),
          privilege_object('SSO', generate_crud_privileges(:sso)),
          privilege_object('Data Connect', data_connect_privileges),
          privilege_object('GraphQL API Explorer', graphql_explorer_privileges),
          privilege_object('Service Integration', service_integration_privileges)
        ]
      end

      def integration_privileges
        [
            privilege_object('Ansible', ansible_privileges),
            privilege_object('Chef', chef_privileges),
            privilege_object('Datadog', datadog_privileges),
            privilege_object('Integration', integrations_privileges),
            privilege_object('New Relic', newrelic_privileges),
            privilege_object('Wavefront', wavefront_privileges),
            privilege_object('Dynatrace', dynatrace_privileges),
            privilege_object('Docker', docker_privileges),
            privilege_object('Puppet', puppet_privileges),
            privilege_object('PagerDuty', pagerduty_privileges),
            privilege_object('Alert Logic', alertlogic_account_privileges),
            privilege_object('Slack', slack_privileges)
        ]
      end

      def services_privileges
        [
          privilege_object('Cost Reallocation', [
              privilege_object('Rulesets',
                generate_udc_privileges(:cost_reallocation_ruleset)),
              privilege_object('Rules',
                generate_udc_privileges(:cost_reallocation_rule)),
            ]),
          privilege_object('Containers Cost Distribution', [
              privilege_object('Rulesets',
                               generate_udc_privileges(:containers_cost_distribution_ruleset))
          ]),
          privilege_object('Custom Line Item Rules', custom_line_items_privileges),
          privilege_object('Budgets', generate_udc_privileges(:budget)),
          privilege_object('Migration Quotes', generate_udc_privileges(:migration_quote))
        ]
      end

      def custom_line_items_privileges
        [
            PrivilegeTuple.new(:create, :chargeback_rule, 'Create Custom Line Item Rule'),
            PrivilegeTuple.new(:read, :chargeback_rule, 'Read Custom Line Item Rule'),
            PrivilegeTuple.new(:update, :chargeback_rule, 'Update Custom Line Item Rule'),
            PrivilegeTuple.new(:delete, :chargeback_rule, 'Delete Custom Line Item Rule'),
        ]
      end

      def other_privileges
        [
          privilege_object('Activity Feed', activity_feed_privileges),
          privilege_object('Custom Dashboards', generate_udc_privileges(:dashboard)),
          privilege_object('Data Export Job', data_export_job_privileges),
          privilege_object('Notifications', system_notification_privileges),
          privilege_object('Efficiency Targets', efficiency_target_privileges, hide_feature?('azure_sql_db_rightsizing')),
          privilege_object('AWS Account Assignments', aws_account_assignment_privileges, true, true),
          privilege_object('EDP', edp_privileges, true),
          privilege_object('Preferences and Schedules', viz_privileges, true), # deprecated
          privilege_object('Cost Savings', cost_savings_privileges, true),
          privilege_object('Savings Automator',icm_privileges),
          privilege_object('Webhook', webhook_privileges),
          privilege_object('Webhook Event Type', webhook_event_type_privileges),

        ]
      end

      def governance_privileges
        [
          privilege_object('Policies', policies_privileges),
          privilege_object('Automated Actions', delegated_actions_privileges),
          privilege_object('Tag Management', tag_management_privileges),
          privilege_object('Anomaly Detection', cost_anomaly_detection_privileges),
          privilege_object('Forecast Snapshot', forecast_snapshot_privileges),
        ]
      end

      def create_customer_privilege
        [PrivilegeTuple.new(:create, :customer)]
      end

      def customer_privileges
        generate_crud_privileges(:customer)
      end

      def owned_customer_privileges
        [
          PrivilegeTuple.new(:read_owned, :customer),
          PrivilegeTuple.new(:update_owned, :customer),
          PrivilegeTuple.new(:delete_owned, :customer),
        ]
      end

      def activity_feed_privileges
        [
          PrivilegeTuple.new(:read, :activity_feed),
          PrivilegeTuple.new(:read_rss, :activity_feed),
          PrivilegeTuple.new(:subscribe, :activity_feed),
        ]
      end

      def system_notification_privileges
        [
          PrivilegeTuple.new(:read, :system_notification),
          PrivilegeTuple.new(:delete, :system_notification),
        ]
      end

      def efficiency_target_privileges
        [
          generate_crud_privileges(:custom_efficiency_targets)
        ].flatten(1)
      end

      def msp_privileges
        [
          generate_crud_privileges(:msp_client),
          generate_crud_privileges(:aws_credit),
          [
            PrivilegeTuple.new(:pinning, :aws_credit)
          ]
        ].flatten(1)
      end

      def aggregator_privileges
        [
          generate_crud_privileges(:aggregator),
          generate_asset_privileges(:aggregator),
          [
            PrivilegeTuple.new(:upload_logs, :aggregator, 'Upload Aggregator Logs'),
            PrivilegeTuple.new(:disable, :aggregator),
            PrivilegeTuple.new(:create, :aggregator_token),
          ]
        ].flatten(1)
      end

      def data_connect_privileges
        [ generate_crud_privileges(:data_connect),
          generate_data_connect_privileges(:data_connect)
        ].flatten(1)
      end

      def profile_privileges
        [
          PrivilegeTuple.new(:read, :profile),
          PrivilegeTuple.new(:update, :profile),
          PrivilegeTuple.new(:generate_api_key, :profile),
        ]
      end

      def tag_management_privileges
        [
          PrivilegeTuple.new(:read, :tag_management),
          PrivilegeTuple.new(:update, :tag_management),
        ]
      end

      def cost_anomaly_detection_privileges
        [
          PrivilegeTuple.new(:read, :cost_anomaly_detection, 'Read Anomaly'),
          PrivilegeTuple.new(:update, :cost_anomaly_detection, 'Update Anomaly'),
          PrivilegeTuple.new(:export, :cost_anomaly_detection, 'Export Anomaly'),
        ]
      end

      def forecast_snapshot_privileges
        [
          PrivilegeTuple.new(:create, :forecast_snapshot, 'Create Forecast Snapshot'),
          PrivilegeTuple.new(:read, :forecast_snapshot, 'Read Forecast Snapshot'),
          PrivilegeTuple.new(:update, :forecast_snapshot, 'Update Forecast Snapshot'),
          PrivilegeTuple.new(:delete, :forecast_snapshot, 'Delete Forecast Snapshot'),
        ]
      end

      def users_privileges
        [
          generate_crud_privileges(:user),
          generate_asset_privileges(:user),
          generate_crud_privileges(:user_to_iam_user),
          [
            PrivilegeTuple.new(:bulk_invite, :user),
            PrivilegeTuple.new(:self_edit_roles, :user, 'Edit Own Role'),
            PrivilegeTuple.new(:self_edit_organization, :user, 'Edit Own Organization'),
            PrivilegeTuple.new(:self_edit, :user),
            # Intentionally removed because with
            # role documents and role bindings the
            # concept of "switching" contexts
            # is baked into the system and is not
            # a permission that is given to a user.
            #
            # PrivilegeTuple.new(:switch_organization, :user, 'Switch Organizations'),
            # PrivilegeTuple.new(:switch_role, :user, 'Switch Roles'),
          ]
        ].flatten(1)
      end

      def containers_privileges
        [
          generate_crud_privileges(:containers_cluster),
          generate_privileges([:read, :update], :containers_configuration),
        ].flatten(1)
      end

      def saved_report_privileges
        generate_udc_privileges_struct(:custom_report, :custom_report_subscription)
      end

      def viz_privileges
        generate_udc_privileges(:viz_preference, :viz_schedule)
      end

      def cost_savings_privileges
        [
          PrivilegeTuple.new(:create, :improvement_tracking),
          PrivilegeTuple.new(:read, :improvement_tracking),
          PrivilegeTuple.new(:tag, :improvement_tracking),
          PrivilegeTuple.new(:export, :improvement_tracking),
          PrivilegeTuple.new(:ignore, :improvement_asset),
          PrivilegeTuple.new(:read, :improvement_asset)
        ]
      end

      def icm_privileges
        [
            PrivilegeTuple.new(:read, :icm, "Read Savings Automator"),
            PrivilegeTuple.new(:create,:icm, "Create Savings Automator"),
            PrivilegeTuple.new(:update,:icm, "Update Savings Automator")
        ]
      end

      def rightsizing_privileges
        [
            PrivilegeTuple.new(:read, :rightsizing_recommendation,'Read Rightsizing Recommendations'),
            PrivilegeTuple.new(:read, :rightsizing_efficiency_target,'Read Rightsizing Efficiency Targets'),
            PrivilegeTuple.new(:create,:rightsizing_efficiency_target,'Create Rightsizing Efficiency Targets'),
            PrivilegeTuple.new(:update,:rightsizing_efficiency_target,'Update Rightsizing Efficiency Targets'),
            PrivilegeTuple.new(:delete,:rightsizing_efficiency_target,'Delete Rightsizing Efficiency Targets')
        ]
      end

      def ansible_privileges
        [
          privilege_object('Accounts', generate_crud_privileges(:ansible_account)),
          privilege_object('Assets', generate_asset_privileges_struct(:ansible_node))
        ]
      end

      def chef_privileges
        [
          privilege_object('Accounts', chef_account_privileges),
          privilege_object('Assets', generate_asset_privileges_struct(:chef_environment,
              :chef_node, :chef_cookbook, :chef_cookbook_version))
        ]
      end

      def chef_account_privileges
        generate_crud_privileges(:chef_account) + [
          PrivilegeTuple.new(:scan, :chef_account),
          PrivilegeTuple.new(:clear_data, :chef_account)
        ]
      end

      def datadog_privileges
        [
            privilege_object('Accounts', generate_crud_privileges(:datadog_account)),
            privilege_object('Assets', generate_asset_privileges_struct(:datadog_host)),
            privilege_object('Assets', generate_asset_privileges_struct(:datadog_tag))
        ]
      end

      def integrations_privileges
        [
            privilege_object('Accounts', generate_crud_privileges(:integration_account)),
            privilege_object('Assets', generate_asset_privileges_struct(:integration_node)),
            privilege_object('Assets', generate_asset_privileges_struct(:integration_tag))
        ]
      end

      def newrelic_privileges
        [
            privilege_object('Accounts', generate_crud_privileges(:newrelic_account)),
            privilege_object('Assets', generate_asset_privileges_struct(:newrelic_server)),
            privilege_object('Assets', generate_asset_privileges_struct(:newrelic_application))
        ]
      end

      def wavefront_privileges
        [
            privilege_object('Accounts', generate_crud_privileges(:wavefront_account)),
            privilege_object('Assets', generate_asset_privileges_struct(:wavefront_source))
        ]
      end

      def dynatrace_privileges
        [
          privilege_object('Accounts', generate_crud_privileges(:dynatrace_account)),
          privilege_object('Assets', generate_asset_privileges_struct(:dynatrace_host))
        ]
      end

      def jira_privileges
        [
            privilege_object('Accounts', generate_crud_privileges(:jira_account)),
        ]
      end

      def slack_privileges
        [
            privilege_object('Accounts', generate_crud_privileges(:slack_account)),
        ]
      end

      def puppet_privileges
        [
          privilege_object('Accounts', puppet_account_privileges),
          privilege_object('Assets', generate_asset_privileges_struct(:puppet_class, :puppet_node)),
        ]
      end

      def puppet_account_privileges
        generate_crud_privileges(:puppet_account) + [
          PrivilegeTuple.new(:scan, :puppet_account),
          PrivilegeTuple.new(:clear_data, :puppet_account),
        ]
      end

      def pagerduty_privileges
        [
          privilege_object('Accounts', pagerduty_account_privileges),
        ]
      end

      def pagerduty_account_privileges
        generate_crud_privileges(:pagerduty_account, :pagerduty_incident) + [
          PrivilegeTuple.new(:scan, :pagerduty_account),
          PrivilegeTuple.new(:clear_data, :pagerduty_account),
        ]
      end

      def docker_privileges
        [
          privilege_object('Assets', generate_asset_privileges(:docker_image, :docker_container)),
        ]
      end

      def policies_privileges
        [
          generate_udc_privileges(:policy),
          privilege_object('Alert Configs (deprecated)', generate_crud_privileges(:alert_config), true),
          [
            PrivilegeTuple.new(:copy, :policy),
            PrivilegeTuple.new(:enable, :policy),
            PrivilegeTuple.new(:disable, :policy),
            PrivilegeTuple.new(:enable_owned, :policy),
            PrivilegeTuple.new(:disable_owned, :policy),
            privilege_object('Policy Notifications', [
                PrivilegeTuple.new(:read, :policy_notification),
                PrivilegeTuple.new(:delete, :policy_notification),
                PrivilegeTuple.new(:resolve, :policy_notification),
                PrivilegeTuple.new(:resolve_owned, :policy_notification),
                PrivilegeTuple.new(:comment, :alert),
                PrivilegeTuple.new(:delete, :alert)]),
          ],
        ].flatten(1)
      end

      def delegated_actions_privileges
        [
          privilege_object('Action Configuration', [
              generate_crud_privileges(:action_config), [
                PrivilegeTuple.new(:enable, :action_config),
                PrivilegeTuple.new(:disable, :action_config),
              ]
            ].flatten(1)),
          privilege_object('Task Configurations', task_config_privileges),
          privilege_object('Automated Actions', [
            PrivilegeTuple.new(:create, :delegated_action, 'Create Automated Action'),
            PrivilegeTuple.new(:update, :delegated_action, 'Update  Automated Action'),
            PrivilegeTuple.new(:read, :delegated_action, 'Read Automated Action'),
            PrivilegeTuple.new(:comment, :delegated_action, 'Comment on Automated Actions'),
            PrivilegeTuple.new(:delete, :delegated_action, 'Delete Automated Action'),
            PrivilegeTuple.new(:clear, :delegated_action, 'Clear Automated Action'),
            PrivilegeTuple.new(:approve, :delegated_action, 'Approve Automated Action'), # approve/deny
            PrivilegeTuple.new(:authorize, :delegated_action, 'Authorize Automated Action'), # authorize/deny
            PrivilegeTuple.new(:cancel, :delegated_action, 'Cancel Automated Action'),
          ])
        ]
      end

      def ir_drilldown_cost_privileges
        [
          PrivilegeTuple.new(:read, :ir_drilldown_cost, "Cost Drilldowns"),
        ]
      end

      def graphql_explorer_privileges
        [
          PrivilegeTuple.new(:read, :graphql_explorer, "GraphQL API Explorer"),
        ]
      end

      def service_integration_privileges
        [
          PrivilegeTuple.new(:read, :service_integration, "Read Service Integration"),
          PrivilegeTuple.new(:approve, :service_integration, "Approve Service Integration"),
          PrivilegeTuple.new(:revoke, :service_integration, "Revoke Service Integration")
        ]
      end

      #############################################
      # AWS Privileges
      #############################################
      def aws_privileges
        [
          privilege_object('Dashboard', dashboard_privileges(:aws)),
          privilege_object('Reports', [
              privilege_object('Pulses', aws_pulse_privileges),
              privilege_object('Cost', aws_cost_report_privileges),
              privilege_object('Usage', aws_usage_report_privileges),
              privilege_object('Performance', aws_performance_report_privileges),
              privilege_object('Metrics', aws_metrics_report_privileges),
              privilege_object('Metrics Drilldown', metrics_data_privileges),
              privilege_object('Security', aws_security_report_privileges),
              privilege_object('Audit', aws_audit_report_privileges),
              privilege_object('Rightsizing', aws_rightsizing_report_privileges),
            ]),
          privilege_object('Assets', aws_asset_privileges),
          privilege_object('Accounts', aws_account_privileges),
          privilege_object('RI Services', ri_service_privileges),
          privilege_object('Savings Plans', aws_savings_plan_privileges),
          privilege_object('Agents', agent_privileges(:aws)),
        ]
      end

      def aws_savings_plan_privileges
        [
            PrivilegeTuple.new(:purchase, :savings_plan_persistence, "Purchase Savings Plan" ),
            PrivilegeTuple.new(:read, :savings_plan_recommendations, "Read Savings Plan Recommendations" )
        ].flatten(1)
      end

      def aws_account_privileges
        [
          generate_crud_privileges(:aws_account),
          generate_asset_privileges(:aws_account),
          [
            PrivilegeTuple.new(:scan, :aws_account),
            PrivilegeTuple.new(:tag, :aws_account),
            PrivilegeTuple.new(:import, :aws_account),
          ]
        ].flatten(1)
      end

      def aws_pulse_privileges
        [
          # Cost Pulse
          privilege_object('Cost', [
              PrivilegeTuple.new(:read, :cost_pulse),
              PrivilegeTuple.new(:subscribe, :cost_pulse),
              PrivilegeTuple.new(:update, :cost_pulse)
            ]),

          # RI Utilization Pulse
          privilege_object('RI', [
              PrivilegeTuple.new(:read, :ri_utilization_pulse),
              PrivilegeTuple.new(:subscribe, :ri_utilization_pulse),
            ]),

          # Health Check Pulse
          privilege_object('Health Check', [
              PrivilegeTuple.new(:read, :health_check_pulse),
              PrivilegeTuple.new(:subscribe, :health_check_pulse),
            ])
        ]
      end

      def aws_cost_report_privileges
        cost_report_privileges = []
        cost_report_privileges << aws_cost_history_privileges
        cost_report_privileges + generate_report_privileges_struct(
          ['History (Original)', :cost_history_original_report, hide_feature?('chargebacks_enabled')],
          ['Current', :cost_current_report],
          ['Savings', :cost_savings_report],
          ['Billing Rule Savings', :billing_rule_savings],
          ['Amortized', :cost_amortized_report],
          ['Budget vs. Actual', :budget_vs_actual_report],
          ['EC2 Instance', :instance_cost_report],
          ['ElastiCache Cost', :ecache_cost_report],
          ['Redshift Cost', :redshift_cost_report],
          ['EBS Volume', :volume_cost_report],
          ['EC2 RI Amortization', :ec2_amortization_report],
          ['RDS RI Amortization', :rds_amortization_report],
          ['ElastiCache Amortization', :ecache_amortization_report],
          ['Redshift Amortization', :redshift_amortization_report],
          ['S3', :s3_cost_report],
          ['Data Transfer', :aws_dxfer_cost],
          ['RDS', :rds_cost_report],
          ['DynamoDB', :dynamo_db_cost_report],
          ['AWS Statement', :aws_statement],
          ['AWS Statement Item', :aws_statement_item])
      end

      def aws_cost_history_privileges
        cost_history_child = []
        cost_history_child << generate_privilege(REPORT_PRIVILEGES, :cost_history_report)
        cost_history_child << aws_cost_history_measure_privileges
        privilege_object("History", cost_history_child.flatten!)
      end

      def aws_cost_history_measure_privileges
        privilege_object('Measures',
                         [
                             PrivilegeTuple.new(:read, :aws_measure_cost_before_chargeback, 'Show "Cost before Chargeback ($)"'),
                         ])
      end

      def aws_usage_report_privileges
        generate_report_privileges_struct(
          ['EC2 Instance', :instance_usage_report],
          ['EC2 Instance Hours', :aws_instance_usage_hours_monthly],
          ['DynamoDB Tables', :aws_dynamo_db_table_usage_hours_monthly],
          ['EMR Cluster Hours', :aws_emr_cluster_usage_hours_monthly],
          ['EBS Volume Hours', :aws_volume_usage_hours_monthly],
          ['EBS Volume', :volume_usage_report],
          ['EC2 RI Inventory', :ec2_ri_inventory_report],
          ['EC2 RI Utilization', :ec2_ri_utilization_report],
          ['EC2 Underutilized RIs', :aws_unused_reservation],
          ['EC2 Snapshot', :ec2_snapshot_report],
          ['RDS Instance', :rds_instance_report],
          ['RDS Instance Hours', :aws_rds_instance_usage_hours_monthly],
          ['RDS RI Inventory', :rds_ri_inventory_report],
          ['RDS RI Utilization', :rds_ri_utilization_report],
          ['RDS Snapshot', :rds_snapshot_report],
          ['ElastiCache', :ecache_usage_report],
          ['ElastiCache RN Inventory ', :ecache_reservation_inventory],
          ['Elasticsearch', :aws_elasticsearch_usage_hours_monthly],
          ['Redshift Usage', :redshift_usage_report],
          ['Redshift RN Inventory', :redshift_reservation_inventory],
          ['S3 Usage', :s3_usage_report],
          ['Data Transfer', :aws_dxfer_usage],
          ['Cross Account AZ Usage', :cross_account_az_usage_report],
          ['WorkSpace Hours', :aws_workspace_usage_hours_monthly])
      end

      def aws_performance_report_privileges
        generate_report_privileges_struct(
          ['EC2 Perspective Rightsizing', :aws_perspective_instance_rightsizing],
          ['EBS Perspective Rightsizing', :aws_perspective_volume_rightsizing],
          ['EBS Rightsizing', :aws_volume_metric_with_cost],
          ['EC2 Rightsizing', :aws_instance_metric_with_cost],
          ['EC2 Instance', :instance_performance_report],
          ['EBS Volume', :volume_performance_report],
          ['EC2 Network', :network_performance_report],
          ['EC2 Local Disk', :disk_performance_report],
          ['EC2 File System', :filesystem_performance_report],
          ['EC2 Memory', :memory_performance_report],
          ['Load Balancer', :load_balancer_performance_report],
          ['RDS Instance', :rds_instance_performance_report])
      end

      def aws_metrics_report_privileges
        generate_report_privileges_struct(
          ['EBS Volume', :aws_volume_cloud_watch_metric],
          ['EC2 File System', :instance_fs_utilization],
          ['EC2 Memory', :instance_mem_utilization],
          ['Load Balancer', :aws_load_balancer_metric],
          ['RDS Instance', :aws_rds_instance_metric],
          ['Instance', [:aws_instance_metric, :aws_instance_cloud_watch_metric]],
          ['S3', [:s3_storage_report_metric, :s3_enhanced_metric]],
          ['DynamoDB', :aws_dynamo_db_table_metric])
      end

      def aws_security_report_privileges
        [
          privilege_object('Open Ports', generate_report_privileges(:aws_security_report)),
          privilege_object('Alert Logic Report', generate_report_privileges(:alertlogic_incidents))
        ]
      end

      def aws_audit_report_privileges
        generate_report_privileges_struct(
          ['Audit', :audit_record])
      end

      def aws_rightsizing_report_privileges
        [
            privilege_object('AWS EC2 Rightsizing Reports', generate_udc_privileges(:aws_ec2_rightsizing_reports))
        ]
      end

      def aws_asset_privileges
        [
          generate_taggable_asset_privileges_struct(
            :aws_auto_scaling_group,
            :aws_cloud_formation_stack,
            :aws_cloudtrail_trail,
            :aws_dedicated_host,
            :aws_dms_replication_instance,
            :aws_dynamo_db_table,
            :aws_elastic_file_system,
            :aws_elasticsearch_domain,
            :aws_emr_cluster,
            :aws_image,
            :aws_instance,
            :aws_kinesis_firehose,
            :aws_kinesis_stream,
            :aws_lambda_function,
            :aws_load_balancer,
            :aws_redshift_cluster,
            :aws_rds_instance,
            :aws_rds_subnet_group,
            :aws_rds_snapshot,
            :aws_route53_hosted_zone,
            :aws_s3_bucket,
            :aws_security_group,
            :aws_snapshot,
            :aws_spot_request,
            :aws_volume,
            :aws_vpc,
            :aws_vpc_subnet,
            :aws_workspace,
            :cloud_front_distribution,
            :aws_sage_maker_endpoint,
            :aws_sagemaker_notebook_instance,
            :aws_sqs_queue,
            :aws_data_pipeline_pipeline,
            :aws_direct_connect_connection,
            :aws_direct_connect_virtual_interface,
            :aws_cognito_user_pool,
            :aws_cloudhsm_hsm,
            :aws_glacier_vault,
            :aws_sagemaker_training_job,
            :aws_cloud_directory_directory,
            :aws_cloud_search_domain,
            :aws_kinesis_video_stream,
            :aws_route53_health_check,
            :aws_code_build_project,
            :aws_dynamo_db_accelerator_cluster,
            :aws_machine_learning_batch_prediction,
            :aws_machine_learning_datasource,
            :aws_machine_learning_evaluation,
            :aws_machine_learning_ml_model,
            :aws_ec2_network_load_balancer,
            :aws_ec2_application_load_balancer,
            :aws_api_gateway_stage,
            :aws_directory_service_directory,
            :aws_alexa_for_business_device,
            :aws_alexa_for_business_user,
            :aws_app_stream_image_builder,
            :aws_app_stream_fleet,
            :aws_iot_analytics_channel,
            :aws_iot_analytics_pipeline,
            :aws_iot_analytics_datastore,
            :aws_iot_analytics_dataset,
            :aws_storage_gateway_file_share,
            :aws_storage_gateway_tape,
            :aws_storage_gateway_volume_gateway,
            :aws_secrets_manager_secret,
            :aws_glue_development_endpoint,
            :aws_glue_job,
            :aws_glue_crawler,
            :aws_lex_bot,
            :aws_cm_certificate,
            :aws_cm_certificate_authority,
            :aws_shield_subscription,
            :aws_shield_protection,
            :aws_waf_rule,
            :aws_waf_webacl,
            :aws_simple_email_service_dedicated_ip,
            :aws_opsworks_server,
            :aws_app_stream_user,
            :aws_gamelift_fleet,
            :aws_route53_traffic_policy,
            :aws_elemental_media_live_input,
            :aws_elemental_media_live_channel,
            :aws_pinpoint_app,
            :aws_kinesis_analytics_application,
            :aws_mq_broker,
            :aws_elemental_media_store_container,
            :aws_greengrass_device,
            :aws_snowball_job,
            :aws_virtual_private_cloud_vpc_endpoint,
            :aws_virtual_private_cloud_vpn_connection,
            :aws_eks_cluster,
            :aws_iot_device_defender_device,
            :aws_sumerian_binary,
            :aws_msk_cluster,
            :aws_ecr_repository,
            :aws_neptune_cluster,
            :aws_neptune_instance,
            :aws_chime_dial_in_conference,
            :aws_chime_account,
            :aws_iot_1_click_device,
            :aws_datasync_task,
            :aws_rds_instance_reservation,
            :aws_instance_reservation,
            :aws_cache_cluster,
            :aws_workspace_bundle,
            :aws_nat_gateway,
            :aws_firewall_manager_policy,
            :aws_amplify_app,
            :aws_quicksight_user,
            :aws_fsx_backup,
            :aws_fsx_file_system,
            :aws_step_functions_state_machine,
            :aws_global_accelerator_accelerator,
            :aws_dynamo_db_backup,
            :aws_transfer_for_sftp_server,
            :aws_cost_management_savings_plan,
            :aws_athena_workgroup,
            :aws_aurora_cluster,
            :aws_aurora_cluster_snapshot,
            :aws_cloud_map_service,
            :aws_data_exchange_data_set,
            :aws_document_db_cluster,
            :aws_document_db_instance,
            :aws_document_db_snapshot,
            :aws_ec2_capacity_reservation,
            :aws_ecs_fargate_task,
            :aws_eks_fargate_pod,
            :aws_elastic_ip,
            :aws_elemental_media_connect_flow,
            :aws_elemental_media_connect_output,
            :aws_elemental_media_connect_source,
            :aws_elemental_media_convert_job,
            :aws_elemental_media_convert_job_template,
            :aws_kendra_data_source,
            :aws_kendra_index,
            :aws_managed_blockchain_member,
            :aws_managed_blockchain_node,
            :aws_sage_maker_batch_transform_job,
            :aws_sage_maker_processing_job,
            :aws_sage_maker_studio_notebook,
            :aws_security_hub_hub,
            :aws_eks_node_group,
            :aws_rds_security_group,
            :aws_transcribe_transcription_job,
            :aws_bedrock_model
            # aws_asset_generator_taggable
          ),
          generate_asset_privileges_struct(
            :aws_ads_connection,
            :aws_ads_process,
            :aws_ads_server,
            :aws_availability_zone,
            :aws_cloud_formation_resource,
            :aws_config_setting,
            :aws_config,
            :aws_elasticache_reserved_node,
            :aws_elasticsearch_instance,
            :aws_elasticsearch_volume,
            :aws_emr_cluster_instance_group,
            :aws_emr_cluster_instance,
            :aws_group,
            :aws_iam_credential_report,
            :aws_iam_password_policy,
            :aws_iam_policy,
            :aws_iam_role,
            :aws_iam_server_certificate,
            :aws_instance_reservation_listing,
            :aws_instance_reservation_modification_item,
            :aws_instance_reservation_modification,
            :aws_instance_status,
            :aws_instance_type,
            :aws_kms_cmk,
            :aws_rds_security_group_rule,
            :aws_redshift_reserved_node,
            :aws_redshift_snapshot,
            :aws_region,
            :aws_security_group_rule,
            :aws_tag,
            :aws_usage,
            :aws_user,
            :cloud_front_origin,
            :instance_file_system,
            :aws_cloudwatch_alarm,
            :aws_elasticsearch_reserved_instance
            # aws_asset_generator_not_taggable
          ),
          [
            PrivilegeTuple.new(:retrieve, :aws_usage),
            PrivilegeTuple.new(:delete, :aws_usage),
            PrivilegeTuple.new(:clear_data, :aws_usage),
            PrivilegeTuple.new(:clear_data, :aws_statement),
            PrivilegeTuple.new(:scan, :aws_statement),
            PrivilegeTuple.new(:modify, :aws_instance_reservation),
            PrivilegeTuple.new(:modify, :aws_rds_instance_reservation),

            PrivilegeTuple.new(:read, :aws_config_rule, 'Read AWS Config Rules', true),
            PrivilegeTuple.new(:export, :aws_config_rule, 'Export AWS Config Rules', true),
            PrivilegeTuple.new(:save, :aws_config_rule, 'Save AWS Config Rules', true),
            PrivilegeTuple.new(:subscribe, :aws_config_rule, 'Subscribe to AWS Config Rules', true),

            PrivilegeTuple.new(:read, :aws_config_rule_evaluation_result, 'Read Config Evaluation Results', true),
            PrivilegeTuple.new(:export, :aws_config_rule_evaluation_result, 'Export Config Evaluation Results', true),
            PrivilegeTuple.new(:save, :aws_config_rule_evaluation_result, 'Save Config Evaluation Results', true),
            PrivilegeTuple.new(:subscribe, :aws_config_rule_evaluation_result, 'Subscribe to Config Evaluation Results', true),

            PrivilegeTuple.new(:comment, :aws_instance_status),
            PrivilegeTuple.new(:delete, :aws_instance_status)
          ],
          generate_privileges([:create, :update, :delete], :aws_statement),
        ].flatten(1)
      end

      def alertlogic_account_privileges
          generate_crud_privileges(:alertlogic_account)
      end

      def ri_service_privileges
        [
          privilege_object('RDS RI Optimizer', [
              generate_udc_privileges(:aws_rds_ri_quote),
              [PrivilegeTuple.new(:purchase, :aws_rds_ri_quote)]
            ].flatten(1)),
          privilege_object('EC2 RI Modifier', [
            PrivilegeTuple.new(:export, :instance_reallocation),
            PrivilegeTuple.new(:read, :instance_reallocation)
            ]),
          privilege_object('EC2 RI Optimizer', [
              generate_udc_privileges(:aws_ri_modeler),
              [PrivilegeTuple.new(:purchase, :aws_ri_modeler)]
            ].flatten(1)),
          privilege_object('EC2 RI Analyzer', [
              PrivilegeTuple.new(:read, :aws_reservation_usage),
              PrivilegeTuple.new(:export, :aws_reservation_usage),
            ]),
          privilege_object('EC2 Convertible RI Exchanger', [
              PrivilegeTuple.new(:read, :aws_ec2_convertible_exchange, 'Read EC2 Convertible RI Exchanger Results'),
              PrivilegeTuple.new(:execute, :aws_ec2_convertible_exchange, 'Execute EC2 Convertible RI Exchanger Results', true)
            ])
        ]
      end

      def metrics_data_privileges
        [
          PrivilegeTuple.new(:read, :metrics_data)
        ]
      end

      #############################################
      # GCP Privileges
      #############################################
      def gcp_privileges
        [
          privilege_object('Dashboard', dashboard_privileges(:gcp)),
          privilege_object('Reports', gcp_report_privileges),
          privilege_object('Assets', gcp_asset_privileges),
          privilege_object('Projects', gcp_project_privileges),
          privilege_object('Billing Accounts', gcp_billing_account_privileges),
          privilege_object('Agents', agent_privileges(:gcp)),
          privilege_object('Pulse', gcp_pulse_privileges),
          privilege_object('GCE Rightsizing', gcp_rightsizing_privileges),
        ]
      end

      def gcp_asset_privileges
        [
          generate_taggable_asset_privileges_struct(
            :gcp_compute_instance,
            :gcp_compute_disk,
            :gcp_compute_image,
            :gcp_compute_snapshot,
            :gcp_big_table_instance,
            :gcp_big_table_cluster,
            :gcp_big_table_table,
            :gcp_big_table_backup,
            :gcp_cloud_run_service,
            :gcp_compute_router,
            :gcp_big_query_table,
            :gcp_big_query_dataset,
            :gcp_compute_cloud_function,
            :gcp_dns_managed_zone,
            :gcp_kms_keyring,
            :gcp_kms_cryptokey,
            :gcp_kms_cryptokey_version,
            :gcp_compute_commitment,
            :gcp_compute_vpn_gateway,
            :gcp_app_engine_application,
            :gcp_sql_database_instance,
            :gcp_app_engine_service,
            :gcp_app_engine_service_version,
            :gcp_logging_logbucket,
            :gcp_secret_manager_secret,
            :gcp_secret_manager_secret_version,
            :gcp_compute_instance_group,
            :gcp_compute_instance_template
            ),
          generate_asset_privileges_struct(
            :gcp_storage_bucket,
            :gcp_dataproc_cluster,
            :gcp_gke_cluster,
            :gcp_billing_statement_item,
            :gcp_billing_statement,
            :gcp_compute_attached_disk,
            :gcp_compute_disk_type,
            :gcp_compute_disk_usage_daily,
            :gcp_compute_image_usage_daily,
            :gcp_compute_instance_usage_daily,
            :gcp_compute_machine_type,
            :gcp_compute_region,
            :gcp_compute_snapshot_usage_daily,
            :gcp_compute_zone,
            :gcp_tag,
            :gcp_usage_statement_item,
            :gcp_usage_statement,
            :gcp_compute_static_ip,
          ),
        ].flatten(1)
      end

      def gcp_report_privileges
        cost_report_privileges = []
        cost_report_privileges << gcp_cost_history_invoice_privileges
        cost_report_privileges << gcp_gce_usage_report_privileges
        cost_report_privileges << gcp_bq_usage_report_privileges
        cost_report_privileges << gcp_bq_cost_report_privileges
        cost_report_privileges + generate_report_privileges_struct(
          :gcp_cost_current_report,
          :gcp_cost_history_report,
          :gcp_compute_engine_cost_report,
          ['Gcp Cost History Invoice (Original) Report', :gcp_cost_history_invoice_original_report, hide_feature?('chargebacks_enabled')],
          :gcp_cost_history_report_v2,
          :gcp_cost_history_invoice_report_v2
        )
      end

      def gcp_gce_usage_report_privileges
        privilege_object('Gcp Gce Usage Report', generate_privileges(GCP_GCE_USAGE_REPORT_PRIVILEGES, :gcp_gce_usage_report))
      end

      def gcp_bq_usage_report_privileges
        privilege_object('Gcp BQ Usage Report', generate_privileges(GCP_BQ_USAGE_REPORT_PRIVILEGES, :gcp_bq_usage_report))
      end

      def gcp_bq_cost_report_privileges
        privilege_object('Gcp BQ Cost Report', generate_privileges(GCP_BQ_COST_REPORT_PRIVILEGES, :gcp_bq_cost_report))
      end

      def gcp_cost_history_invoice_privileges
        cost_history_invoice_child = [
            generate_privilege(REPORT_PRIVILEGES, :gcp_cost_history_invoice_report),
            gcp_cost_history_invoice_measure_privileges
        ]
        privilege_object("Gcp Cost History Invoice Report", cost_history_invoice_child.flatten!)
      end

      def gcp_cost_history_invoice_measure_privileges
        privilege_object('Measures',
                         [
                             PrivilegeTuple.new(:read, :gcp_measure_cost_before_chargeback, 'Show "Cost before Chargeback ($)"'),
                         ])
      end

      def gcp_pulse_privileges
        [
            # Cost Pulse
            privilege_object('GCP Cost Pulse', [
                PrivilegeTuple.new(:read, :gcp_cost_pulse),
                PrivilegeTuple.new(:subscribe, :gcp_cost_pulse),
                PrivilegeTuple.new(:update, :gcp_cost_pulse)
            ]),
        ]
      end

      def gcp_project_privileges
        [
          generate_crud_privileges(:gcp_compute_project),
          generate_taggable_asset_privileges(:gcp_compute_project),
        ].flatten(1)
      end

      def gcp_billing_account_privileges
        [
            generate_crud_privileges(:gcp_billing_account),
            generate_taggable_asset_privileges(:gcp_billing_account),
            [
                PrivilegeTuple.new(:scan, :gcp_billing_account),
            ]
        ].flatten(1)
      end

      def gcp_rightsizing_privileges
        [
            generate_crud_privileges(:gcp_gce_rightsizing_recommendation),
            generate_crud_privileges(:gcp_gce_rightsizing_operation)
        ].flatten(1)
      end
      #############################################
      # VMWARE Privileges
      #############################################
      def vmware_privileges
        [
          privilege_object('Dashboard', dashboard_privileges(:vmware)),
          privilege_object('Assets', vmware_asset_privileges),
          privilege_object('Accounts', vmware_account_privileges),
          privilege_object('Agents', agent_privileges(:vmware)),
        ]
      end

      def vmware_asset_privileges
        generate_asset_privileges_struct(:vmware_datacenter,
          :vmware_folder,
          :vmware_host,
          :vmware_datastore,
          :vmware_virtual_machine)
      end

      def vmware_account_privileges
        [
          generate_crud_privileges(:vmware_account),
          generate_taggable_asset_privileges(:vmware_account),
          [
            PrivilegeTuple.new(:scan, :vmware_account),
            PrivilegeTuple.new(:clear_data, :vmware_account),
          ]
        ].flatten(1)
      end

      #############################################
      # DATACENTER Privileges
      #############################################
      def data_center_privileges
        # 'datacenter' is not a typo, due to how this is dynamically constructed in code
        [
          privilege_object('Dashboard', dashboard_privileges(:datacenter)),
          privilege_object('Reports', [
            privilege_object('Cost',       data_center_cost_report_privileges),
            privilege_object('Usage',       data_center_usage_report_privileges),
            privilege_object('Performance', data_center_performance_report_privileges),
            privilege_object('Metrics', data_center_metrics_report_privileges),
            privilege_object('Audit',       data_center_audit_report_privileges),
          ]),
          privilege_object('Assets', data_center_asset_privileges),
          privilege_object('Accounts', data_center_account_privileges),
          privilege_object('Machine', data_center_server_privileges),
          privilege_object('Hosts', data_center_host_privileges),
          privilege_object('DataStores', data_center_data_store_privileges),
          privilege_object('Clusters', data_center_cluster_privileges),
          privilege_object('DataCenters', data_center_data_center_privileges),
          privilege_object('Agents', agent_privileges(:datacenter)),
          privilege_object('PricingProfile', data_center_pricing_profile_privileges),
          privilege_object('PricingConfiguration', data_center_pricing_configuration_privileges)
        ]
      end

      def data_center_asset_privileges
        [
          generate_asset_privileges(
            :data_center_tag,
            :data_center_server_file_system,
            :data_center_server_network_interface,
            :data_center_server_block_device,
            :data_center_server_cpu,
            :data_center_server_usage_daily
          )
        ].flatten(1)
      end

      def data_center_usage_report_privileges
        generate_report_privileges_struct(['Machine', :data_center_server_usage_report])
      end

      def data_center_cost_report_privileges
        generate_report_privileges_struct(
            ['Machine', :data_center_server_cost_report],
            ['Current', :data_center_cost_current_report],
            ['Cost Driver', :data_center_cost_driver_report]
        )
      end

      def data_center_performance_report_privileges
        generate_report_privileges_struct(
          ['Machine',       :data_center_server_performance_report],
          ['File System',  :data_center_filesystem_performance_report],
          ['Memory',       :data_center_memory_performance_report],
          ['Network',      :data_center_network_performance_report],
          ['Block Device', :data_center_block_device_performance_report],
          ['Data Center Server Rightsizing', :data_center_server_rightsizing]
        )
      end

      def data_center_metrics_report_privileges
        generate_report_privileges_struct(
          ['Machine',       :data_center_server_metric],
          ['File System',  :data_center_server_file_system_utilization],
          ['Memory',       :data_center_server_mem_utilization],
          ['Network',      :data_center_server_network_utilization],
          ['Block Device', :data_center_server_block_device_utilization],
        )
      end

      def data_center_audit_report_privileges
        generate_report_privileges_struct(['Audit', :data_center_audit_record])
      end

      def data_center_account_privileges
        [
          generate_crud_privileges(:data_center_account),
          generate_taggable_asset_privileges(:data_center_account),
          [
            PrivilegeTuple.new(:upload_payload, :data_center_account),
            PrivilegeTuple.new(:scan, :data_center_account),
            PrivilegeTuple.new(:clear_data, :data_center_account),
          ]
        ].flatten(1)
      end

      def data_center_server_privileges
        [
          [PrivilegeTuple.new(:update, :data_center_server)], # For updating data center server pricing
          generate_taggable_asset_privileges(:data_center_server)
        ].flatten(1)
      end

      def data_center_host_privileges
        generate_taggable_asset_privileges(:data_center_host)
      end

      def data_center_data_store_privileges
        generate_taggable_asset_privileges(:data_center_data_store)
      end

      def data_center_cluster_privileges
        generate_taggable_asset_privileges(:data_center_cluster)
      end

      def data_center_data_center_privileges
        generate_taggable_asset_privileges(:data_center_data_center)
      end

      def data_center_pricing_profile_privileges
        [
          generate_crud_privileges(:data_center_pricing_profile),
          generate_taggable_asset_privileges(:data_center_pricing_profile)
        ].flatten(1)
      end

      def data_center_pricing_configuration_privileges
        [
          generate_crud_privileges(:data_center_server_configuration)
        ].flatten(1)
      end

      #############################################
      # VMC Privileges
      #############################################
      def vmware_cloud_privileges
        [
          privilege_object('Dashboard', dashboard_privileges(:vmc)),
          privilege_object('Reports', [
                              privilege_object('Cost', vmware_cloud_cost_report_privileges),
                              privilege_object('Usage', vmware_cloud_usage_report_privileges),
                              ]),
          privilege_object('Assets', vmc_asset_privileges),
          privilege_object('Accounts', vmc_vcenter_privileges)
        ]
      end

      def vmc_vcenter_privileges
        [
            generate_crud_privileges(:vmc_vcenter),
            generate_taggable_asset_privileges(:vmc_vcenter),
            [ PrivilegeTuple.new(:upload_payload, :vmc_vcenter),
              PrivilegeTuple.new(:scan, :vmc_vcenter),
              PrivilegeTuple.new(:clear_data, :vmc_vcenter),
            ]
        ].flatten(1)
      end

      def vmware_cloud_cost_report_privileges
        generate_report_privileges_struct(
          ['History', :vmc_cost_history_report],
          ['Current Billing Period', :vmc_cost_current_billing_period_report],
          ['Previous Billing Period', :vmc_cost_previous_billing_period_report]
        )
      end

      def vmware_cloud_usage_report_privileges
        generate_report_privileges_struct(
            ['Host Hours', :vmc_usage_host_report],
            ['Elastic IP Hours', :vmc_usage_elastic_ip_report],
            ['Data Transfer', :vmc_usage_data_transfer_report]
        )
      end

      def vmc_asset_privileges
        [
          generate_taggable_asset_privileges_struct(
            :vmc_sddc,
            :vmc_data_center,
            :vmc_cluster,
            :vmc_host,
            :vmc_machine,
          ),
          generate_crud_privileges(
            :vmware_csp_organization
          ),
          generate_asset_privileges(
            :vmware_csp_organization,
            :vmc_tag
          ),
          [
            PrivilegeTuple.new(:scan, :vmware_csp_organization),
          ],
          [
            PrivilegeTuple.new(:tag, :vmware_csp_organization),
          ]
        ].flatten(1)
      end

      #############################################
      # Partner reports Privileges
      #############################################
      def partner_privileges
        [
          privilege_object('MSPs', msp_privileges),
          privilege_object('AWS Account Assignments', aws_account_assignment_privileges),
          privilege_object('GCP Billing Account Assignments', gcp_billing_account_assignment_privileges),
          privilege_object('Partner Billing Rule', partner_billing_rule_privileges),
          privilege_object('Customer AWS Statements', customer_statement_privileges),
          privilege_object('Reports', partner_reports_privileges),
          privilege_object('Custom Price Books', price_books_privileges),
          privilege_object('Partner Lock Day', partner_lock_day_privileges, hide_feature?('partner_policy')),
        ]
      end

      def partner_reports_privileges
        generate_report_privileges_struct(:partner_cost_history_report,
                                          :partner_margin_history_report)
      end

      def price_books_privileges
        [
            generate_crud_privileges(:price_book),
            generate_crud_privileges(:price_book_assignment),
        ].flatten(1)
      end

      def partner_lock_day_privileges
        [
          PrivilegeTuple.new(:create, :partner_lock_day),
          PrivilegeTuple.new(:read, :partner_lock_day),
          PrivilegeTuple.new(:update, :partner_lock_day)
        ]
      end

      #############################################
      # AZURE Privileges
      #############################################
      def azure_privileges
        # 'datacenter' is not a typo, due to how this is dynamically constructed in code
        [
            privilege_object('Dashboard', dashboard_privileges(:azure)),
            privilege_object('Reports', [
                privilege_object('Pulse', azure_pulse_privileges),
                privilege_object('Cost', azure_cost_report_privileges),
                privilege_object('Usage', azure_usage_report_privileges)
                ]),
            privilege_object('Assets', azure_asset_privileges),
            privilege_object('Agents', agent_privileges(:azure)),
            privilege_object('Enrollments', azure_enrollment_privileges),
            privilege_object('Performance', azure_performance_report_privileges),
            privilege_object('Metrics', azure_metrics_report_privileges),
            privilege_object('Azure RI Services', azure_ri_service_privileges),
            privilege_object('Azure Rightsizing Reports', azure_rightsizing_reports_privileges, hide_feature?('azure_sql_db_rightsizing')),
            privilege_object('Azure Partner Customer Accounts', azure_partner_customer_accounts_privileges),
            privilege_object('Azure Account Assignments', azure_account_assignment_privileges)
        ]
      end

      def azure_asset_privileges
        [
            generate_taggable_asset_privileges(:azure_vm),
            generate_taggable_asset_privileges(:azure_vm_scale_set),
            generate_taggable_asset_privileges(:azure_storage),
            generate_taggable_asset_privileges(:azure_resource_group),
            generate_taggable_asset_privileges(:azure_ip_address),
            generate_taggable_asset_privileges(:azure_security_group),
            generate_taggable_asset_privileges(:azure_security_rule),
            generate_taggable_asset_privileges(:azure_sql_server),
            generate_taggable_asset_privileges(:azure_sql_database),
            generate_taggable_asset_privileges(:azure_sql_managed_instance),
            generate_taggable_asset_privileges(:azure_redis_cache),
            generate_taggable_asset_privileges(:azure_virtual_network_gateway),
            generate_taggable_asset_privileges(:azure_app_service_plan),
            generate_taggable_asset_privileges(:azure_service_bus_namespace),
            generate_taggable_asset_privileges(:azure_virtual_network),
            generate_taggable_asset_privileges(:azure_batch_account),
            generate_taggable_asset_privileges(:azure_backup_vault),
            generate_taggable_asset_privileges(:azure_recovery_services_vault),
            generate_taggable_asset_privileges(:azure_hd_insight_cluster),
            generate_taggable_asset_privileges(:azure_search_service),
            generate_taggable_asset_privileges(:azure_sql_data_warehouse),
            generate_taggable_asset_privileges(:azure_stor_simple_device_manager),
            generate_taggable_asset_privileges(:azure_app_service),
            generate_taggable_asset_privileges(:azure_log_analytics_workspace),
            generate_taggable_asset_privileges(:azure_express_route_circuit),
            generate_taggable_asset_privileges(:azure_cdn_profile),
            generate_taggable_asset_privileges(:azure_snapshot),
            generate_taggable_asset_privileges(:azure_network_interface),
            generate_taggable_asset_privileges(:azure_managed_disk),
            generate_taggable_asset_privileges(:azure_app_insight),
            generate_taggable_asset_privileges(:azure_document_db),
            generate_taggable_asset_privileges(:azure_key_vault),
            generate_taggable_asset_privileges(:azure_application_gateway),
            generate_taggable_asset_privileges(:azure_app_service_environment),
            generate_taggable_asset_privileges(:azure_reservation_order),
            generate_taggable_asset_privileges(:azure_reservation),
            generate_taggable_asset_privileges(:azure_cloud_service),
            generate_taggable_asset_privileges(:azure_log_profile),
            generate_taggable_asset_privileges(:azure_activity_log_alert),
            generate_taggable_asset_privileges(:azure_key_vault_key),
            generate_taggable_asset_privileges(:azure_key_vault_secret),
            generate_taggable_asset_privileges(:azure_subscription),
            generate_taggable_asset_privileges(:azure_kubernetes_service),
            generate_taggable_asset_privileges(:azure_iot_hub),
            generate_taggable_asset_privileges(:azure_billing_account),
            generate_taggable_asset_privileges(:azure_billing_profile),
            generate_taggable_asset_privileges(:azure_invoice_section),
            generate_taggable_asset_privileges(:azure_event_hub_namespace),
            generate_taggable_asset_privileges(:azure_event_hub),
            generate_taggable_asset_privileges(:azure_policy_assignment),
            generate_taggable_asset_privileges(:azure_policy_state),
            generate_taggable_asset_privileges(:azure_postgresql_server),
            generate_taggable_asset_privileges(:azure_databricks_workspace),
            generate_taggable_asset_privileges(:azure_reservation_v2),
            generate_taggable_asset_privileges(:azure_mysql_server),
            generate_taggable_asset_privileges(:azure_savings_plan),
            generate_taggable_asset_privileges(:azure_waf_application_gateway_policy),
            generate_taggable_asset_privileges(:azure_avs_private_cloud),
            generate_taggable_asset_privileges(:azure_machine_learning_private_endpoint_connection),
            generate_taggable_asset_privileges(:azure_machine_learning_compute_resource),
            generate_taggable_asset_privileges(:azure_machine_learning_workspace),
            generate_taggable_asset_privileges(:azure_compute_availability_set),
            generate_taggable_asset_privileges(:azure_container_registry_registry),
            generate_taggable_asset_privileges(:azure_avs_cluster),
            generate_taggable_asset_privileges(:azure_avs_datastore),
            generate_taggable_asset_privileges(:azure_avs_virtual_machine),
            generate_asset_privileges(
                :azure_subscription,
                :azure_account,
                :azure_department,
                :azure_cost_center,
                :azure_tag,
                :azure_zone,
                :azure_region,
                :azure_vm_file_system,
                :azure_vm_network_interface,
                :azure_vm_virtual_disk,
                :azure_vm_cpu,
                :azure_vm_size,
                :azure_vm_rate,
                :azure_catalog,
                :azure_cloud_service,
                :azure_vm_reservation_rate,
                :azure_sql_database_size,
                :azure_storage_blob_container,
                :azure_vm_instance,
                :azure_key_vault_diagnostic_setting,
                :azure_sql_server_firewall_rule,
                :azure_subscription_security_policy,
                :azure_sql_database_auditing,
                :azure_sql_server_auditing,
                :azure_sql_server_threat_detection,
                :azure_sql_database_threat_detection,
                :azure_active_directory_user,
                :azure_active_directory_role_definition,
                :azure_active_directory_role_definition_action,
                :azure_sql_server_vulnerability_assessment,
                :azure_vm_spot_info,
                :azure_postgresql_server_firewall_rule,
            )
        ].flatten(1)
      end

      def azure_pulse_privileges
        [
            # Cost Pulse
            privilege_object('Cost', [
                PrivilegeTuple.new(:read, :azure_cost_pulse),
                PrivilegeTuple.new(:subscribe, :azure_cost_pulse),
                PrivilegeTuple.new(:update, :azure_cost_pulse)
            ]),
            privilege_object('Azure Health Check', [
                PrivilegeTuple.new(:read, :azure_health_check_pulse),
                PrivilegeTuple.new(:subscribe, :azure_health_check_pulse),
                PrivilegeTuple.new(:update, :azure_health_check_pulse)
            ])
        ]
      end

      def azure_cost_report_privileges
        cost_report_privileges = []
        cost_report_privileges << azure_cost_history_privileges
        cost_report_privileges + generate_report_privileges_struct(
            ['Burndown', :azure_cost_burndown_report],
            ['Cost (Original)', :azure_cost_history_original_report, hide_feature?('chargebacks_enabled')],
            ['Current', :azure_cost_current_report],
            ['Current Billing Period', :azure_cost_current_billing_period_report],
            ['Previous Billing Period', :azure_cost_previous_billing_period_report],
            ['Amortized', :azure_cost_amortized_report],
            ['VM', :azure_vm_cost_report],
            ['Disk', :azure_disk_cost_report],
            ['Snapshot', :azure_snapshot_cost_report],
            ['SQL Database', :azure_sql_database_cost_report],
            ['Storage_Account', :azure_storage_account_cost_report]
        )
      end

      def azure_cost_history_privileges
        cost_history_child = [
            generate_privilege(REPORT_PRIVILEGES, :azure_cost_history_report),
            azure_cost_history_measure_privileges
        ]
        privilege_object("Cost", cost_history_child.flatten!)
      end

      def azure_cost_history_measure_privileges
        privilege_object('Measures',
                         [
                             PrivilegeTuple.new(:read, :azure_measure_cost_before_chargeback, 'Show "Cost before Chargeback ($)"'),
                         ])
      end

      def azure_usage_report_privileges
        generate_report_privileges_struct(
          ['VM', :azure_vm_usage_report],
          ['VM_HOURS', :azure_vm_usage_hours_monthly],
          ['SQL_DB_HOURS', :azure_sql_db_usage_hours_monthly],
          ['Disk', :azure_disk_usage_report],
          ['Snapshot', :azure_snapshot_usage_report],
          ['SQL Database', :azure_sql_database_usage_report],
          ['VM_RI_Inventory', :azure_vm_ri_inventory_report],
          ['VM_RI_Utilization', :azure_vm_ri_utilization_report],
          ['VM_Underutilized', :azure_vm_underutilized_report],
          ['Storage_Account', :azure_storage_account_usage_report]
        )
      end

      def azure_enrollment_privileges
        [
            generate_crud_privileges(:azure_enrollment),
            generate_taggable_asset_privileges(:azure_enrollment),
            [
                PrivilegeTuple.new(:scan, :azure_enrollment),
                PrivilegeTuple.new(:clear_data, :azure_enrollment),
            ],
            generate_crud_privileges(:azure_service_principal),
            generate_taggable_asset_privileges(:azure_service_principal),
            [
                PrivilegeTuple.new(:scan, :azure_service_principal),
                PrivilegeTuple.new(:clear_data, :azure_service_principal),
            ],
            generate_crud_privileges(:azure_partner_center),
            generate_taggable_asset_privileges(:azure_partner_center),
            [
                PrivilegeTuple.new(:scan, :azure_partner_center),
                PrivilegeTuple.new(:clear_data, :azure_partner_center),
            ]
        ].flatten(1)
      end

      def azure_performance_report_privileges
        generate_report_privileges_struct(
            ['Virtual Machine', :azure_vm_performance_report],
            ['SQL Database', :azure_sql_database_performance_report],
            ['File System',  :azure_filesystem_performance_report],
            ['Memory',       :azure_memory_performance_report],
            ['Network',      :azure_network_performance_report],
            ['Virtual Disk', :azure_virtual_disk_performance_report],
            ['Azure Rightsizing by VM', :azure_vm_rightsizing],
            ['Azure Perspective VM Rightsizing', :azure_perspective_vm_rightsizing]
        )
      end

      def azure_metrics_report_privileges
        generate_report_privileges_struct(
            ['Virtual Machine', :azure_vm_metric],
            ['File System',  :azure_vm_file_system_utilization],
            ['Memory',       :azure_vm_mem_utilization],
            ['Network',      :azure_vm_network_utilization],
            ['Virtual Disk', :azure_vm_virtual_disk_utilization],
            ['SQL Database', :azure_sql_database_metric],
        )
      end

      def azure_ri_service_privileges
        [
          privilege_object('Azure VM RI Optimizer',
            generate_udc_privileges(:azure_vm_ri_quote)
          )
        ]
      end

      def azure_rightsizing_reports_privileges
        [
          privilege_object('Azure SQL Database Rightsizing Reports',
            generate_udc_privileges(:azure_sql_db_rightsizing_reports)
          )
        ]
      end

      def azure_partner_customer_accounts_privileges
        generate_privileges(AZURE_PARTNER_CUSTOMER_ACCOUNT_PRIVILEGES, :azure_partner_customer)
      end

      def azure_account_assignment_privileges
        [
            privilege_object('Azure Account Assignments',
                             generate_crud_privileges(:azure_account_assignment)
            )
        ]
      end

      #############################################
      # Multi Cloud Privileges
      #############################################
      def multi_cloud_privileges
        [
            privilege_object('Reports', [
                privilege_object('Cost', multi_cloud_cost_report_privileges)
            ])
        ]
      end

      def multi_cloud_cost_report_privileges
        generate_report_privileges_struct(
            ['History', :multi_cloud_cost_history_report]
        )
      end

      #############################################
      # Multidimensional Reports Privileges
      #############################################
      def multidimensional_reports_privileges
        [
            PrivilegeTuple.new(:save, :multidimensional_reports)
        ]
      end

      #############################################
      # FlexReport Privileges
      #############################################

      def flex_report_privileges
        flex_report_privilege_list = [
          generate_udc_privileges(:flex_report),
          privilege_object('Data Sources', flex_dataset_privileges)
        ]
        flex_report_privilege_list.flatten!
      end

      #############################################
      # Dataset Privileges
      #############################################

      def flex_dataset_privileges
        [
          privilege_object('AWS Datasets', generate_aws_dataset_privileges),
          privilege_object('Azure Datasets', generate_azure_dataset_privileges),
          privilege_object('GCP Datasets', generate_gcp_dataset_privileges),
          privilege_object('Kubernetes Datasets', generate_kubernetes_dataset_privileges),
          privilege_object('Multicloud Datasets', generate_multicloud_dataset_privileges)
        ]
      end

      def generate_aws_dataset_privileges
        [
          PrivilegeTuple.new(:read, 'aws_cur_flex_dataset'.to_sym, "AWS Cost & Usage Report"),
          PrivilegeTuple.new(:read, 'aws_cost_anomaly_flexreport_dataset'.to_sym, "AWS Anomaly Flex Report")
        ]
      end

      def generate_azure_dataset_privileges
        [
          PrivilegeTuple.new(:read, 'azure_ea_flex_dataset'.to_sym, "Azure Enterprise Agreement"),
          PrivilegeTuple.new(:read, 'azure_cost_anomaly_flexreport_dataset'.to_sym, "Azure Anomaly Flex Report")
        ]
      end

      def generate_gcp_dataset_privileges
        [
          PrivilegeTuple.new(:read, 'gcp_billing_export_flex_dataset'.to_sym, "GCP BigQuery Billing Export"),
          PrivilegeTuple.new(:read, 'gcp_cost_anomaly_flexreport_dataset'.to_sym, "GCP Anomaly Flex Report")
        ]
      end

      def generate_kubernetes_dataset_privileges
        [
          PrivilegeTuple.new(:read, 'kubernetes_containers_usage_flexreport_dataset'.to_sym, "Kubernetes Containers Usage"),
          PrivilegeTuple.new(:read, 'kubernetes_nodes_usage_flexreport_dataset'.to_sym, "Kubernetes Nodes Usage"),
          PrivilegeTuple.new(:read, 'azure_kubernetes_cost_dataset'.to_sym, "Azure Kubernetes Cost"),
          PrivilegeTuple.new(:read, 'aws_kubernetes_cost_dataset'.to_sym, "AWS Kubernetes Cost")
        ]
      end

      def generate_multicloud_dataset_privileges
        [
          PrivilegeTuple.new(:read, 'multicloud_carbon_emissions_flexreport_dataset'.to_sym, "MultiCloud Operational Emissions")
        ]
      end

      #############################################
      # Container related privileges
      #############################################
      def container_report_privileges
        report_privileges = generate_report_privileges_struct(
          ['Kubernetes Resources',    :container_cluster_kubernetes],
          ['ECS Resources',           :container_cluster_ecs],
          ['Kubernetes Allocation',   :container_utilization_kubernetes],
          ['ECS Allocation',          :container_utilization_ecs],
          ['Containers Cost History', :container_cost_history],
          ['Azure Containers Cost History', :azure_container_cost_history])
        report_privileges << privilege_object('Kubernetes Containers Usage',
                                              generate_privileges(K8S_REPORT_PRIVILEGES,
                                                                  :kubernetes_containers_usage_view))
        report_privileges << privilege_object('Kubernetes Nodes Usage',
                                              generate_privileges(K8S_REPORT_PRIVILEGES,
                                                                  :kubernetes_nodes_usage_view))
        report_privileges << privilege_object('AWS Kubernetes Cost History',
                                              generate_privileges(K8S_REPORT_PRIVILEGES,
                                                                  :aws_kubernetes_cost_view))
        report_privileges << privilege_object('Azure Kubernetes Cost History',
                                              generate_privileges(K8S_REPORT_PRIVILEGES,
                                                                  :azure_kubernetes_cost_view))
        report_privileges

      end


      #############################################
      # EDP Privileges
      #############################################
      def edp_privileges
        [
          generate_crud_privileges(:aws_billing_contract),
          generate_crud_privileges(:aws_billing_rule),
          generate_crud_privileges(:aws_billing_rule_pricing_plan),
          generate_crud_privileges(:aws_billing_rule_price_point)
        ].flatten(1)
      end
      #############################################
      # Partner Billing Rule Privileges
      #############################################
      def partner_billing_rule_privileges
        [
          generate_crud_privileges(:partner_billing_rule),
          generate_crud_privileges(:aws_partner_support_billing_rule),
          generate_crud_privileges(:partner_custom_billing_rule)
        ].flatten(1)
      end

      def aws_account_assignment_privileges
        generate_crud_privileges(:aws_account_assignment)
      end

      def gcp_billing_account_assignment_privileges
        generate_crud_privileges(:gcp_billing_account_assignment)
      end

      def task_config_privileges
        generate_crud_privileges(:task_config)
      end

      def customer_statement_privileges
        generate_crud_privileges(:customer_statement)
      end

      #############################################
      # ORACLE Privileges
      #############################################
      def oracle_privileges
        [
            privilege_object('Tenancies', [
                generate_crud_privileges(:oci_tenancy),
                generate_taggable_asset_privileges(:oci_tenancy),
                [
                    PrivilegeTuple.new(:scan, :oci_tenancy),
                    PrivilegeTuple.new(:clear_data, :oci_tenancy),
                ]
            ].flatten(1)
            ),
            privilege_object('Dashboard', dashboard_privileges(:oci)),
            privilege_object('Reports', [
                privilege_object('Cost', oracle_cost_report_privileges),
                privilege_object('Usage', oracle_usage_report_privileges)
            ]),
            privilege_object('Assets', oracle_asset_privileges)
        ]
      end


      def oracle_cost_report_privileges
        generate_report_privileges_struct(
            ['Cost', :oci_cost_history_report]
        )
      end

      def oracle_usage_report_privileges
        generate_report_privileges_struct(
            ['VM', :oci_vm_usage_report]
        )
      end

      def oracle_asset_privileges
        [
            generate_taggable_asset_privileges(:oci_vm),
            generate_taggable_asset_privileges(:oci_compartment),
            generate_privileges(OCI_GENERIC_ASSET_PRIVILEGES,:oci_generic_asset),
            generate_privileges(OCI_REGION_PRIVILEGES,:oci_region),
        ].flatten(1)
      end

      #############################################
      # ALIBABA Privileges
      #############################################
      #
      def alicloud_privileges
        [
          privilege_object('Dashboard', dashboard_privileges(:alicloud)),
          privilege_object('Reports', [
            privilege_object('Cost', alicloud_cost_history_report),
          ]),
          privilege_object('Assets', alicloud_asset_privileges)
        ]
      end

      def alicloud_cost_history_report
        generate_report_privileges_struct(
          ['Cost', :alicloud_cost_history_report]
        )
      end
      # #placeholder for alibaba_assets
      def alicloud_asset_privileges
        [
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_generic_asset),
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_region),
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_accounts),
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_cost_center),
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_elastic_block_storages),
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_elastic_compute_services),
          generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES,:alicloud_server_load_balancers)
        ].flatten(1)
      end

      #############################################
      # KUBERNETES Privileges
      #############################################

      def kubernetes_privileges
        [
            privilege_object('Assets', kubernetes_asset_privileges),
        ]
      end

      def kubernetes_asset_privileges
        # Since containers_cluster is already added as part of containers_privileges, adding only missing privileges
        [
            PrivilegeTuple.new(:export, :"containers_cluster"),
            PrivilegeTuple.new(:save,  :"containers_cluster"),
            PrivilegeTuple.new(:subscribe, :"containers_cluster"),
        ].concat(generate_asset_privileges_struct(:kubernetes_namespace,
                                                  :kubernetes_node,
                                                  :kubernetes_deployment,
                                                  :kubernetes_daemonset,
                                                  :kubernetes_replicaset,
                                                  :kubernetes_statefulset,
                                                  :kubernetes_job,
                                                  :kubernetes_containers_image))
      end

      #############################################
      # Recommendations Privileges
      #############################################
      def recommendation_privileges
        [
          privilege_object('Commitment Discounts', commitment_discounts_privileges),
          privilege_object('Rightsizing',rightsizing_privileges)
        ]
      end

      #############################################
      # Helpers and Generators
      #############################################
      def agent_privileges(cloud)
        [
          PrivilegeTuple.new(:read, :"#{cloud}_agent_configuration"),
          PrivilegeTuple.new(:update, :"#{cloud}_agent_configuration"),
          PrivilegeTuple.new(:disable_autoupdate, :"#{cloud}_agent_status"),
          PrivilegeTuple.new(:enable_autoupdate, :"#{cloud}_agent_status"),
        ].concat(generate_asset_privileges(:"#{cloud}_agent_status"))
      end

      def commitment_discounts_privileges
          [
            PrivilegeTuple.new(:read, :"commitment_discounts")
          ]
      end

      def dashboard_privileges(cloud)
        [
          PrivilegeTuple.new(:read, :"#{cloud}_dashboard"),
          PrivilegeTuple.new(:change, :"#{cloud}_dashboard")
        ]
      end

      def data_export_job_privileges
        [
          PrivilegeTuple.new(:read, :export_job),
          PrivilegeTuple.new(:create, :export_job),
          PrivilegeTuple.new(:update, :export_job),
          PrivilegeTuple.new(:delete, :export_job),
        ]
      end

      def webhook_privileges
        [
          PrivilegeTuple.new(:read, :webhook),
          PrivilegeTuple.new(:create, :webhook),
          PrivilegeTuple.new(:update, :webhook),
          PrivilegeTuple.new(:delete, :webhook),
        ]
      end

      def webhook_event_type_privileges
        [
          PrivilegeTuple.new(:read, :webhook_event_type),
        ]
      end

      def generate_data_connect_privileges(*subjects)
        generate_privileges(DATA_CONNECT_PRIVILEGES, *subjects)
      end

      def generate_report_privileges_struct(*subjects)
        generate_privileges_struct(REPORT_PRIVILEGES, *subjects)
      end

      def generate_report_privileges(*subjects)
        generate_privileges(REPORT_PRIVILEGES, *subjects)
      end

      def generate_asset_privileges_struct(*subjects)
        generate_privileges_struct(ASSET_PRIVILEGES, *subjects)
      end

      def generate_asset_privileges(*subjects)
        generate_privileges(ASSET_PRIVILEGES, *subjects)
      end

      def generate_taggable_asset_privileges_struct(*subjects)
        generate_privileges_struct(TAGGABLE_ASSET_PRIVILEGES, *subjects)
      end

      def generate_taggable_asset_privileges(*subjects)
        generate_privileges(TAGGABLE_ASSET_PRIVILEGES, *subjects)
      end

      def generate_crud_privileges(*subjects)
        generate_privileges(CRUD_PRIVILEGES, *subjects)
      end

      def generate_udc_privileges_struct(*subjects)
        generate_privileges_struct(CRUD_PRIVILEGES + UDC_PRIVILEGES, *subjects)
      end

      def generate_udc_privileges(*subjects)
        generate_privileges(CRUD_PRIVILEGES + UDC_PRIVILEGES, *subjects)
      end

      def generate_oci_generic_asset_privileges(*subjects)
        generate_privileges(OCI_GENERIC_ASSET_PRIVILEGES, *subjects)
      end

      def generate_alicloud_generic_asset_privileges(*subjects)
        generate_privileges(ALICLOUD_GENERIC_ASSET_PRIVILEGES, *subjects)
      end

      def generate_privileges_struct(actions, *subjects)
        subjects.map do |val|
          title, subs, hidden = if val.kind_of?(Array)
                                  hidden = (val.count == 3) ? val[2] : false
                                  [val.first, val[1], hidden]
                                else
                                  [val.to_s.humanize.titleize, val, false]
                                end
          privilege_object(title, generate_privileges(actions, *subs), hidden)
        end
      end

      def generate_privileges(actions, *subjects)
        subjects.flatten.map{|r| generate_privilege(actions, r)}.flatten(1)
      end

      def generate_privilege(actions, subject)
        raise "Expected Symbol, got #{subject.class.name}" unless subject.kind_of?(::Symbol)
        actions.product(Array(subject)).map{|tuple| PrivilegeTuple.new(*tuple)}
      end

      # If current_customer is available, check if the customer should be allowed
      #   to see privileges for a feature.
      #
      # Return true (hide this feature):
      # - if current_customer isn't within context and we can't determine if we should hide.
      #   (err on the side of not showing the privileges)
      # - if current_customer does not have the feature
      #
      # Return false (show this feature):
      # - if current_customer is within context and the customer does have the
      #   feature defined.
      def hide_feature?(feature)
        # Is current_customer within context?
        customer = current_customer.present? rescue false
        # If current_customer is within context, check if they have the feature.
        # Otherwise, return true to hide the feature
        customer ? !current_customer.has_feature?(feature) : true
      end

      def visit_privilege_nodes(object, &block)
        case object
        when PrivilegeTuple
          yield object
        when PrivilegeSection
          object.children.each{|obj| visit_privilege_nodes(obj, &block)}
        else
          raise "Unexpected type in privilege tree: #{object.inspect}:#{object.class.name}"
        end
      end

      def privilege_nodes(object, nodes = [])
        case object
        when PrivilegeTuple
          nodes + [object]
        when PrivilegeSection
          object.children.inject(nodes){|n, obj| privilege_nodes(obj, n)}
        else
          raise "Unexpected type in privilege tree: #{object.inspect}:#{object.class.name}"
        end
      end
    end
  end
end
