# frozen_string_literal: true

require 'active_record'
require 'import'
require 'role_management/models/content/meta_data'
require 'role_management/errors'
require 'cht-lib-fabric/helpers/wavefront'
require 'hitimes'

module RoleManagement
  # The Stores module contains all the classes and functionality
  # necessary to interact with the datastores that RoleManagement
  # requires. It has very basic ActiveRecord models and a
  # RoleDirectory class that performs the actual role document lookup
  # and authorization.
  module Stores
    class RoleDocument < ::ActiveRecord::Base
      has_many :permissions_role_documents
      has_many :permissions, through: :permissions_role_documents, dependent: :destroy

      has_many :user_group_role_bindings

      def crn
        customer_id = if system_defined?
                        nil
                      else
                        ::Platform::CRNs::Tenant.from_s(self.customer_crn).customer
                      end
        ::RoleManagement::Models::CRNs::RoleDocument.new(customer: customer_id,
                                                         details: self.id.to_s)
      end
    end

    class Permission < ::ActiveRecord::Base
    end

    class PermissionsRoleDocument < ::ActiveRecord::Base
      belongs_to :role_document, touch: true
      belongs_to :permission
    end

    class UserGroupRoleBinding < ::ActiveRecord::Base
      belongs_to :role_document
      has_many :user_group_role_binding_organizations

      def crn
        customer_id = ::Platform::CRNs::Tenant.from_s(self.customer_crn).customer
        ::RoleManagement::Models::CRNs::UserGroupRoleBinding.new(customer: customer_id,
                                                                 role_binding_id: self.id.to_s)
      end
    end

    class UserGroupRoleBindingOrganization < ::ActiveRecord::Base
      belongs_to :user_group_role_binding
    end

    class RoleDocumentsDirectory
      include Import['logger', 'alert', 'wavefront']
      include ::Platform::Helpers::Wavefront

      def stats_by_customer
        {
          role_documents: RoleDocument.where(system_defined: false).group(:customer_crn).count,
          role_bindings: UserGroupRoleBinding.group(:customer_crn).count
        }
      end

      def by_id_and_customer_crn(role_document_id, customer_crn, organization_id, include_permissions: true, service_account: false, include_permission_labels: true)
        wf_time('role_management.role_doc_directory.by_id_and_customer_crn') do
          rel = RoleDocument.where(id: role_document_id,
                                   customer_crn: customer_crn)
          rel = rel.where(organization_id: organization_id) unless service_account
          to_role_doc(rel.first!, include_permissions: include_permissions, include_permission_labels: include_permission_labels)
        end
      end

      def by_id_and_system_defined(role_document_id, include_permissions: true, include_permission_labels: true)
        wf_time('role_management.role_doc_directory.by_id_and_system_defined') do
          rel = RoleDocument.where(id: role_document_id,
                                   system_defined: true)
          to_role_doc(rel.first!, include_permissions: include_permissions, include_permission_labels: include_permission_labels)
        end
      end

      def system_defined(include_permissions: true)
        wf_time('role_management.role_doc_directory.system_defined') do
          rel = RoleDocument.where(system_defined: true)
          rel = rel.includes(:permissions) if include_permissions

          rel.map do |rd|
            to_role_doc(rd, include_permissions: include_permissions)
          end
        end
      end

      def by_customer_crn(customer_crn, organization_id, include_permissions: true, service_account: false)
        wf_time('role_management.role_doc_directory.by_customer_crn') do
          rel = RoleDocument.where(customer_crn: customer_crn)
          rel = rel.where(organization_id: organization_id) unless service_account

          rel.map do |rd|
            to_role_doc(rd, include_permissions: include_permissions)
          end
        end
      end

      def by_shared_role_doc_id(customer_crn, role_document_ids, include_permissions: true, include_permission_labels: true)
        rel = RoleDocument.where(customer_crn: customer_crn,
                                 id: role_document_ids,
                                 system_defined: false)
        rel = rel.includes(:permissions) if include_permissions
        rel.map do |rd|
          # Define content interface values
          metadata = ::RoleManagement::Models::Content::Metadata.new(
            shared: ::RoleManagement::Models::Content::SharedContentMetadata.new(
              enabled: true
            )
          )
          to_role_doc(rd, include_permissions: include_permissions, metadata: metadata, include_permission_labels: include_permission_labels)
        end
      end

      def create_custom_role_document(name:,
                                      description: nil,
                                      restricted:,
                                      customer_crn:,
                                      user_crn:,
                                      permissions: nil,
                                      organization:)
        wf_time('role_management.role_doc_directory.create_custom_role_document') do
          # Use transaction to rollback if permissions are not valid
          role_doc = ::ActiveRecord::Base.transaction do
            role_doc = RoleDocument.create(
              customer_crn: customer_crn,
              created_by_crn: user_crn,
              updated_by_crn: user_crn,
              name: name,
              description: description,
              restricted: restricted,
              organization_id: organization
            )

            if permissions
              role_doc.permissions = valid_permissions(permissions)
            end

            role_doc
          end

          to_role_doc(role_doc)
        end
      end

      def update_custom_role_document(id, customer_crn:,
                                          updated_by_crn:,
                                          name:,
                                          description:,
                                          permissions: nil,
                                          organization:)
        wf_time('role_management.role_doc_directory.update_custom_role_document') do
          role_doc = RoleDocument.where(id: id, customer_crn: customer_crn, organization_id: organization)
                       .first!

          # Accept partial updates of name and description
          updated_name = name || role_doc.name
          updated_description = description || role_doc.description

          # Use transaction to rollback if permissions are not valid
          ::ActiveRecord::Base.transaction do
            role_doc.update(
              updated_by_crn: updated_by_crn,
              name: updated_name,
              description: updated_description
            )

            if permissions
              role_doc.permissions = valid_permissions(permissions)
            end
          end

          to_role_doc(role_doc)
        end
      end

      def destroy_role_document(role_doc_id, customer_crn, organization_id)
        wf_time('role_management.role_doc_directory.destroy_role_document') do
          role_doc = RoleDocument.where(id: role_doc_id,
                                        customer_crn: customer_crn,
                                        system_defined: false,
                                        organization_id: organization_id)
                       .first!
          role_doc.destroy
        end
      end

      def permission_tuple_to_crn
        wf_time('role_management.role_doc_directory.permission_tuple_to_crn') do
          Permission.all.each_with_object({}) do |perm, memo|
            memo[[perm.action.to_sym, perm.subject.to_sym]] =
              ::RoleManagement::Models::CRNs::Permission.new(
                details: perm.id.to_s
              ).to_s
          end
        end
      end

      def user_group_role_binding(
            crn:,
            customer_crn:,
            organization_crn:
          )
        wf_time('role_management.role_doc_directory.user_group_role_binding') do
          to_ugrb UserGroupRoleBinding.where(id: crn.user_group_role_binding_id,
                                             customer_crn: customer_crn.to_s,
                                             organization_crn: organization_crn.to_s)
                    .includes(:user_group_role_binding_organizations)
                    .first!
        end
      rescue ::ActiveRecord::RecordNotFound
        raise RoleManagement::Exceptions::NotFound
      end

      def user_group_role_bindings(
            customer_crn:,
            organization_crn:,
            service_account:,
            include_role_doc_permissions: true,
            include_role_docs: true,
            include_role_bindings: true
          )
        wf_time('role_management.role_doc_directory.user_group_role_bindings') do
          rel = UserGroupRoleBinding.where(customer_crn: customer_crn.to_s)
                  .includes(:user_group_role_binding_organizations)
          # Can we do the same eager / batch loading with an .includes(:role_document)?
          # instead of the batch query and assign?
          rel = rel.where(organization_crn: organization_crn.to_s) unless service_account
          # Fetch all related role documents at once
          # TODO: Check to make sure tenant is honored here, it should
          # be based on the bindings, but compare to rb.role_document generateq query
          role_documents = []
          # Fires the UGRB query
          role_bindings = rel.map
          if include_role_docs
            # Need role document IDS from each binding and fetch them all at once
            # want to avoid looping over UGRB twice if possible, can we do uniq here
            # to avoid fetching the same RD more than once? (i.e same rd in 2 bindings?)
            role_document_ids_from_bindings = role_bindings.map(&:role_document_id)
            # Fires the RD query getting all RD's associated with the all the bindings
            role_documents = RoleDocument.where(id: [role_document_ids_from_bindings]).to_a
            role_documents_by_id = role_documents.each_with_object({}) { |role_document, memo| memo[role_document.id] = role_document }
            # Assign RD to UGRB which means it should not fire again in the to_ugrb method
            role_bindings.each { |role_binding| role_binding.role_document = role_documents_by_id[role_binding.role_document_id] }
          end
          role_bindings.each do |rb|
            # Can we defer the RB lookup here similar to a data loader?
            to_ugrb(rb, include_permissions: include_role_doc_permissions, include_role_docs: include_role_docs, include_role_bindings: include_role_bindings)
          end
        end
      end

      def create_user_group_role_binding(
            customer_crn:,
            user_crn:,
            current_ou_crn:,
            user_group_crn:,
            role_document_crn:,
            organization_crns:
          )

        wf_time('role_management.role_doc_directory.create_user_group_role_binding') do
          org_role_bindings = organization_crns.map do |ou|
            UserGroupRoleBindingOrganization.new(organization_crn: ou.to_s)
          end

          to_ugrb UserGroupRoleBinding.create(
                    user_group_crn: user_group_crn.to_s,
                    role_document_id: role_document_crn.role_document_id,
                    user_group_role_binding_organizations: org_role_bindings,
                    customer_crn: customer_crn.to_s,
                    organization_crn: current_ou_crn.to_s,
                    created_by_crn: user_crn.to_s,
                    updated_by_crn: user_crn.to_s,
                  )
        end
      rescue ActiveRecord::RecordNotUnique => ex
        raise RoleManagement::Exceptions::BindingAlreadyExists
      rescue ActiveRecord::InvalidForeignKey => ex
        # role document is the only foreign key to worry about so it
        # will be the cause of this exception
        raise RoleManagement::Exceptions::InvalidRoleDocument
      end

      def update_user_group_role_binding(
            customer_crn:,
            user_crn:,
            current_ou_crn:,
            ugrb_crn:,
            organization_crns:
          )

        wf_time('role_management.role_doc_directory.update_user_group_role_binding') do
          ugrb = UserGroupRoleBinding.where(customer_crn: customer_crn.to_s,
                                            organization_crn: current_ou_crn.to_s,
                                            id: ugrb_crn.user_group_role_binding_id)
                   .first!
          orgs = organization_crns.map do |ou|
            UserGroupRoleBindingOrganization.new(organization_crn: ou.to_s)
          end
          ugrb.user_group_role_binding_organizations.destroy_all
          ugrb.user_group_role_binding_organizations = orgs

          ugrb.updated_by_crn = user_crn.to_s
          ugrb.touch
          ugrb.save!
          to_ugrb ugrb
        end
      rescue ::ActiveRecord::RecordNotFound
        raise RoleManagement::Exceptions::NotFound
      end

      def delete_user_group_role_binding(
            customer_crn:,
            user_crn:,
            current_ou_crn:,
            ugrb_crn:
          )
        wf_time('role_management.role_doc_directory.delete_user_group_role_binding') do
          ugrb = UserGroupRoleBinding.where(customer_crn: customer_crn.to_s,
                                            organization_crn: current_ou_crn.to_s,
                                            id: ugrb_crn.user_group_role_binding_id)
                   .first!
          ugrb.destroy!
          ugrb_crn.to_s
        end
      rescue ::ActiveRecord::RecordNotFound
        raise RoleManagement::Exceptions::NotFound
      end

      def delete_org_user_group_role_bindings(org_crn:)
        wf_time('role_management.role_doc_directory.delete_org_user_group_role_bindings') do
          UserGroupRoleBinding.where(organization_crn: current_ou_crn.to_s)
            .destroy_all!
        end
      end

      # Destroy all UGRBs and Role Documents associated with an
      # organization.
      # This includes UGRBs created in that organization and
      # also cleaning up UGRBs that reference that organization
      # so that they no longer reference it.
      def delete_organization(org_crn:)
        wf_time('role_management.role_doc_directory.delete_organization') do
          transaction do
            UserGroupRoleBinding.where(organization_crn: org_crn.to_s).destroy_all
            UserGroupRoleBindingOrganization.where(organization_crn: org_crn.to_s).destroy_all
            RoleDocument.where(organization_id: org_crn.organization).destroy_all
          end
        end
      end

      # Fetch all UGRBs associated with the user-group crn
      def user_group_role_bindings_by_user_group(user_group_crn:)
        wf_time('role_management.role_doc_directory.user_group_role_bindings_by_user_group') do
          rel = UserGroupRoleBinding.where(user_group_crn: user_group_crn.to_s)
          rel.map do |rb|
            to_ugrb(rb, include_permissions: false, include_role_docs: true)
          end
        end
      end

      def transaction
        ::ActiveRecord::Base.transaction do
          yield
        end
      end

      def to_ugrb(ugrb, include_permissions: true, include_role_docs: true, include_role_bindings: true)
        return nil unless ugrb

        wf_time('role_management.role_doc_directory.to_ugrb') do
          customer_id = ::Platform::CRN.from_s(ugrb.customer_crn).customer
          orgs = ugrb.user_group_role_binding_organizations.map do |ou|
            ou.organization_crn
          end

          ::RoleManagement::Models::UserGroupRoleBinding.new(
            id:                    ugrb.crn.to_s,
            role_document:         include_role_docs ? to_role_doc(ugrb.role_document, include_permissions: include_permissions, include_role_bindings: include_role_bindings) : nil,
            user_group_id:         ugrb.user_group_crn,
            organization_id:       ugrb.organization_crn,
            organization_ids:      orgs,
            customer_id:           ugrb.customer_crn,
            created_by_id:         ugrb.created_by_crn,
            updated_by_id:         ugrb.updated_by_crn,
            created_at:            ugrb.created_at,
            updated_at:            ugrb.updated_at,
          )
        end
      end

      private

      def valid_permissions(permissions)
        if permissions.count == 0
          # Empty array for permissions so delete them from role doc
          permissions
        else
          # Ensure all permissions are valid and in the db
          valid_permissions = permissions_by_tuples(permissions)
          if valid_permissions.count != permissions.count
            error_message = "Valid Permissions Count [#{valid_permissions.count}] Permissions Count [#{permissions.count}]"
            alert.error error_message
            raise ::RoleManagement::Exceptions::InvalidPermission, error_message
          end
          valid_permissions
        end
      end

      def permissions_by_tuples(permissions)
        action_subj_tuple = permissions.map do |permission|
          [permission["action"], permission["subject"]]
        end
        Permission.where("(action, subject) IN \
                          (#{(['(?)']*action_subj_tuple.size).join(', ')})",
                         *action_subj_tuple)
      end

      def to_role_doc(role_document, include_permissions: true, metadata: nil, include_role_bindings: true, include_permission_labels: true)
        return nil unless role_document
        wf_time('role_management.role_doc_directory.to_role_doc',
                tags: {include_permissions: include_permissions.to_s}) do
          customer_id = ::Platform::CRN.from_s(role_document.customer_crn)
                          .customer

          crn = ::RoleManagement::Models::CRNs::RoleDocument.new(
            customer: customer_id,
            details: role_document.id.to_s
          ).to_s
          ::RoleManagement::Models::RoleDocument.new(
            id:             crn,
            organizationId: role_document.organization_id,
            name:           role_document.name,
            description:    role_document.description,
            created_by:     role_document.created_by_crn,
            updated_by:     role_document.updated_by_crn,
            system_defined: role_document.system_defined,
            restricted:     role_document.restricted,
            created_at:     role_document.created_at,
            updated_at:     role_document.updated_at,
            permissions:    include_permissions ? role_document.permissions.map { |perm| to_permission(perm, include_permission_labels: include_permission_labels) } : [],
            user_group_role_bindings: include_role_bindings ? role_document.user_group_role_bindings : nil,
            metadata:       metadata.nil? ? default_metadata : metadata
          )
        end
      end

      def to_permission(permission, include_permission_labels: true)
        @permission_cache ||= {}
        @permission_cache[permission.id] ||= begin
                                               crn = ::RoleManagement::Models::CRNs::Permission.new(
                                                 details: permission.id.to_s
                                               ).to_s

                                               key = "PrivilegeTuple.#{permission.action}_#{permission.subject}" 
                                               value = include_permission_labels ? Sgtn.t(key, 'cht-svc-role-management') : nil


                                               ::RoleManagement::Models::Permission.new(
                                                 id: crn,
                                                 action:  permission.action,
                                                 subject: permission.subject,
                                                 label: value
                                               )
                                             end
      end

      #
      # Default values for all content
      #
      def default_metadata
        ::RoleManagement::Models::Content::Metadata.new(
          shared: ::RoleManagement::Models::Content::SharedContentMetadata.new(
            enabled: false
          )
        )
      end
    end
  end
end
